package RIO.extaction;
//0702_23:27 Hydrantが0個の場合に対応
//0504_10:30 isPassableを前の行動と比較する方法に変えた
//0504_11:00 isPassableのぬるぽとった
//0504_18:00 calcExtinguishPositionを移行
//0513 calcにおいて，calcExtinguishPositionとcalcExtinguishの順番を逆に
//0513 actionInside追加
//0514 actionInsideにisOnfireの判定を加えた
//0514 actionInsideに本当に視界内にあるのか，Firrynessが定義されているのかの判定を追加
import java.awt.Polygon;
import java.awt.Rectangle;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import adf.agent.action.Action;
import adf.agent.action.common.ActionMove;
import adf.agent.action.fire.ActionExtinguish;
import adf.agent.action.fire.ActionRefill;
import adf.agent.communication.MessageManager;
import adf.agent.communication.standard.bundle.information.MessageFireBrigade;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.communication.CommunicationMessage;
import adf.component.extaction.ExtAction;
import adf.component.module.algorithm.Clustering;
import adf.component.module.algorithm.PathPlanning;
import rescuecore2.config.NoSuchConfigOptionException;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.misc.geometry.Vector2D;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.EntityID;

public class RioneActionFireFighting extends ExtAction {
	private PathPlanning pathPlanning;

	private int maxExtinguishDistance;
	private int maxExtinguishPower;
	private int thresholdRest;
	private int kernelTime;
	private int refillCompleted;
	private int refillRequest;
	//private boolean refillFlag;

	private EntityID target;

	private boolean isRefuge;
	private boolean comunicate;

	private Clustering clustering;


	HashMap<Integer,Action> actionHistory = new HashMap<Integer,Action>();
	HashMap<Integer,EntityID> positionHistory = new HashMap<Integer,EntityID>();
	HashMap<Integer, EntityID> movetargetHistory = new HashMap<Integer,EntityID>();

	public RioneActionFireFighting(AgentInfo agentInfo, WorldInfo worldInfo, ScenarioInfo scenarioInfo, ModuleManager moduleManager, DevelopData developData) {
		super(agentInfo, worldInfo, scenarioInfo, moduleManager, developData);
		this.maxExtinguishDistance = scenarioInfo.getFireExtinguishMaxDistance();
		this.maxExtinguishPower = scenarioInfo.getFireExtinguishMaxSum();
		this.thresholdRest = developData.getInteger("ActionFireFighting.rest", 100);
		this.comunicate = false;
		this.isRefuge = false;

		this.target = null;

		switch  (scenarioInfo.getMode()) {
		case PRECOMPUTATION_PHASE:
			this.pathPlanning = moduleManager.getModule("ActionFireFighting.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
			this.clustering = moduleManager.getModule("SampleBuildingDetector.Clustering", "adf.sample.module.algorithm.SampleKMeans");
			break;
		case PRECOMPUTED:
			this.pathPlanning = moduleManager.getModule("ActionFireFighting.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
			this.clustering = moduleManager.getModule("SampleBuildingDetector.Clustering", "adf.sample.module.algorithm.SampleKMeans");
			break;
		case NON_PRECOMPUTE:
			this.pathPlanning = moduleManager.getModule("ActionFireFighting.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
			this.clustering = moduleManager.getModule("SampleBuildingDetector.Clustering", "adf.sample.module.algorithm.SampleKMeans");
			break;
		}
	}

	@Override
	public ExtAction precompute(PrecomputeData precomputeData) {
		super.precompute(precomputeData);
		if(this.getCountPrecompute() >= 2) {
			return this;
		}
		this.pathPlanning.precompute(precomputeData);
		try {
			this.kernelTime = this.scenarioInfo.getKernelTimesteps();
		}catch (NoSuchConfigOptionException e) {
			this.kernelTime = -1;
		}
		return this;
	}

	@Override
	public ExtAction resume(PrecomputeData precomputeData) {
		super.resume(precomputeData);
		if(this.getCountResume() >= 2) {
			return this;
		}
		this.pathPlanning.resume(precomputeData);
		try {
			this.kernelTime = this.scenarioInfo.getKernelTimesteps();
		}catch (NoSuchConfigOptionException e) {
			this.kernelTime = -1;
		}
		return this;
	}

	@Override
	public ExtAction preparate() {
		super.preparate();
		if(this.getCountPreparate() >= 2) {
			return this;
		}
		this.pathPlanning.preparate();
		try {
			this.kernelTime = this.scenarioInfo.getKernelTimesteps();
		}catch (NoSuchConfigOptionException e) {
			this.kernelTime = -1;
		}
		return this;
	}

	@Override
	public ExtAction updateInfo(MessageManager messageManager){
		super.updateInfo(messageManager);
		int currentTime = this.agentInfo.getTime();
		FireBrigade fireBrigade = (FireBrigade)this.agentInfo.me();
		EntityID agentPosition = fireBrigade.getPosition();
		positionHistory.put(currentTime,agentPosition);

		if(this.getCountUpdateInfo() >= 2) {
			return this;
		}

		for(CommunicationMessage message : messageManager.getReceivedMessageList()) {
			Class<? extends CommunicationMessage> messageClass = message.getClass();
			if(messageClass == MessageFireBrigade.class){
				MessageFireBrigade mfb = (MessageFireBrigade)message;
				if(mfb.getAction() == 2) {
					comunicate = true;
					this.target = mfb.getTargetID();
				}
				//MessageUtil.reflectMessage(this.worldInfo, mfb);
			}
		}

		this.clustering.updateInfo(messageManager);
		this.pathPlanning.updateInfo(messageManager);
		return this;
	}

	@Override
	public ExtAction setTarget(EntityID target) {
		this.target = null;
		if(target != null) {
			StandardEntity entity = this.worldInfo.getEntity(target);
			if(entity instanceof Building) {
				this.target = target;
			}
		}
		return this;
	}

	@Override
	public ExtAction calc() {

		this.result = null;
		FireBrigade agent = (FireBrigade)this.agentInfo.me();
		int agentX = agent.getX();
		int agentY = agent.getY();
		StandardEntity positionEntity = this.worldInfo.getPosition(agent);

		if(positionEntity instanceof Road){
			Road road = (Road)positionEntity;
			if(road.isBlockadesDefined() && road.getBlockades().size() > 0) {
				for(Blockade blockade : this.worldInfo.getBlockades(road)) {
					if(!blockade.isApexesDefined()) {
						continue;
					}
					if(this.isInside(agentX, agentY, blockade.getApexes())) {
						this.result = actionInside();
					}
				}
			}
		}

		if(this.result != null){
			return this;
		}

		this.result = calcRefill();
		if(this.result != null){
			return this;
		}

		this.result = this.calcExtinguish(agent, this.pathPlanning, this.target);
		if(this.result != null){
			return this;
		}

		if(comunicate) {

			for(EntityID id  :worldInfo.getChanged().getChangedEntities()) {
				if(id.getValue()==this.target.getValue()) {
					comunicate = false;

					StandardEntity sEntity = worldInfo.getEntity(this.target);

					if(sEntity instanceof Building) {
						Building building = (Building)sEntity;

						if(building.isOnFire()) {
							this.result = new ActionExtinguish(this.target,this.maxExtinguishPower);
							return this;
						}
					}
				}
			}

			EntityID myId = agentInfo.me().getID();

			//System.out.println("ここ使ってる");

			getMoveAction(pathPlanning, myId, this.target);

		}

		this.result = calcExtinguishPosition(agent, pathPlanning);
		return this;
	}
	private boolean isInside(double pX, double pY, int[] apex) {
		Point2D p = new Point2D(pX, pY);
		Vector2D v1 = (new Point2D(apex[apex.length - 2], apex[apex.length - 1])).minus(p);
		Vector2D v2 = (new Point2D(apex[0], apex[1])).minus(p);
		double theta = this.getAngle(v1, v2);

		StandardEntity se = agentInfo.me();
		Human me = (Human) se;

		if(me.getBuriedness() == 0) {
			return false;
		}

		for(int i = 0; i < apex.length - 2; i += 2) {
			v1 = (new Point2D(apex[i], apex[i + 1])).minus(p);
			v2 = (new Point2D(apex[i + 2], apex[i + 3])).minus(p);
			theta += this.getAngle(v1, v2);
		}
		return Math.round(Math.abs((theta / 2) / Math.PI)) >= 1;
	}

	private double getAngle(Vector2D v1, Vector2D v2) {
		double flag = (v1.getX() * v2.getY()) - (v1.getY() * v2.getX());
		double angle = Math.acos(((v1.getX() * v2.getX()) + (v1.getY() * v2.getY())) / (v1.getLength() * v2.getLength()));
		if(flag > 0) {
			return angle;
		}
		if(flag < 0) {
			return -1 * angle;
		}
		return 0.0D;
	}
	private Action actionInside(){
		Human agent = (Human)this.agentInfo.me();
		StandardEntity positionEntity = this.worldInfo.getPosition(agent);

		//System.out.println("瓦礫の中にいるためactioninside中" + agent.getID());

		if(target != null){
			StandardEntity entity = this.worldInfo.getEntity(target);
			EntityID ei = entity.getID();
			for(EntityID id:worldInfo.getChanged().getChangedEntities()){
				if(entity instanceof Building && id.equals(ei)){
					Building building =(Building)entity;
					if (building.isFierynessDefined() && building.isOnFire() && this.worldInfo.getDistance(positionEntity, entity) < this.maxExtinguishDistance){
						return new ActionExtinguish(target, this.maxExtinguishPower);
					}
				}
			}
			//return new ActionRest();
		}

		return null;

	}

	//追林拓光，西田孝典
	private Action calcRefill(){
		FireBrigade fireBrigade = (FireBrigade)this.agentInfo.me();
		EntityID agentPosition = fireBrigade.getPosition();
		int water = fireBrigade.getWater();//水の量
		StandardEntityURN positionURN = this.worldInfo.getPosition(fireBrigade).getStandardURN();
		int maxWater = scenarioInfo.getFireTankMaximum();
		int Damage = fireBrigade.getDamage();
		boolean hyd_exist = true;

		//Hydrantが無いマップの想定
		if(worldInfo.getEntitiesOfType(StandardEntityURN.HYDRANT).size() == 0){
			hyd_exist = false;
		}

		//Refugeにいる時の処理
		if(positionURN.equals(StandardEntityURN.REFUGE)){
			if(water > maxWater*0.9) { //9割給水していたら
				return null; //何もしない
			}else {
				return new ActionRefill();
			}
		}

		//Hydrantにいる時の処理	
		if(hyd_exist) {
			if(positionURN.equals(StandardEntityURN.HYDRANT)){
				if(water > maxWater*0.4){ //4割給水していたら
					return null; //何もしない
				}else {
					return new ActionRefill();
				}
			}
		}


		//ダメージがある場合
		if(Damage > 0) {
			target = calcBestRefillPosition(hyd_exist, false);
			return this.getMoveAction(pathPlanning, agentPosition, target);
		}

		//給水の必要がある場合
		if(water < maxWater*0.1){
			target = calcBestRefillPosition(hyd_exist, true);
			return this.getMoveAction(pathPlanning, agentPosition, target);
			//TODO ここにHydrantかRefugeかに行く処理を書く
		}

		return null;
	}

	private class DistanceSorter implements Comparator<StandardEntity> {
		private StandardEntity reference;
		private WorldInfo worldInfo;

		DistanceSorter(WorldInfo wi, StandardEntity reference) {
			this.reference = reference;
			this.worldInfo = wi;
		}

		public int compare(StandardEntity a, StandardEntity b) {
			int d1 = this.worldInfo.getDistance(this.reference, a);
			int d2 = this.worldInfo.getDistance(this.reference, b);
			return d1 - d2;
		}
	}

	private EntityID calcBestRefillPosition(boolean hyd_exist , boolean useHY){
		ArrayList<StandardEntity> RefugeList = new ArrayList<>();
		ArrayList<StandardEntity> HydrantList = new ArrayList<>();
		EntityID myID = agentInfo.me().getID();

		StandardEntity myposition = worldInfo.getPosition(myID);
		HashMap<EntityID,Integer> costMap = new HashMap<EntityID,Integer>();
		int limit = 0;

		for(StandardEntity se : worldInfo.getEntitiesOfType(StandardEntityURN.REFUGE)){
			RefugeList.add(se);
		}

		RefugeList.sort(new DistanceSorter(this.worldInfo,agentInfo.me()));

		for(StandardEntity se: RefugeList) {
			int cost = 0;
			EntityID id = se.getID();
			pathPlanning.setFrom(myposition.getID());
			pathPlanning.setDestination(id);

			for(EntityID entityID : pathPlanning.calc().getResult()) {
				StandardEntity se2 = worldInfo.getEntity(entityID);
				if(se2 instanceof Road) {
					Road road = (Road)se2;

					if(road.isBlockadesDefined() && road.getBlockades().size() > 0) {
						for(EntityID bl : road.getBlockades()) {
							Blockade blockade = (Blockade)worldInfo.getEntity(bl);
							cost += blockade.getRepairCost();
						}	
					}
				}
			}
			
			costMap.put(id, cost);
			limit++;
			
			if(limit > 2) {
				break;
			}

		}


		if(useHY) {

			limit = 0;
			
			int index = clustering.getClusterIndex(myID);

			for(StandardEntity se : worldInfo.getEntitiesOfType(StandardEntityURN.HYDRANT)){
				if(clustering.getClusterEntities(index).contains(se)) {
					HydrantList.add(se);
				}
			}
			
			HydrantList.sort(new DistanceSorter(this.worldInfo,agentInfo.me()));

			for(StandardEntity sEntity : HydrantList) {
				int cost = 0;
				pathPlanning.setFrom(myposition.getID());
				pathPlanning.setDestination(sEntity.getID());

				for(EntityID entityID : pathPlanning.calc().getResult()) {
					StandardEntity se = worldInfo.getEntity(entityID);
					if(se instanceof Road) {
						Road road = (Road)se;

						if(road.isBlockadesDefined() && road.getBlockades().size() > 0) {
							for(EntityID bl : road.getBlockades()) {
								Blockade blockade = (Blockade)worldInfo.getEntity(bl);
								cost += blockade.getRepairCost();
							}	
						}
					}
				}

				costMap.put(sEntity.getID(), cost);
				cost = 0;
				
				limit++;
				
				if(limit > 2) {
					break;
				}

			}	
		}

		int min = Integer.MAX_VALUE;
		EntityID refill_position = null;


		if(costMap.size() > 0) {
			for(Map.Entry<EntityID, Integer> entry : costMap.entrySet()) {
				if(entry.getValue() < min) {
					min = entry.getValue();
					refill_position = entry.getKey();
				}
			}

			return refill_position;

		}
		return null;
	}

	private Action calcExtinguishPosition(FireBrigade agent, PathPlanning pathPlanning){
		ArrayList<Building> fireBuildings = new ArrayList<>();//燃えている建物の集まり
		ArrayList<Building> infernoBuildings = new ArrayList<>();//inferno状態の建物の集まり
		ArrayList<Building> burningBuildings = new ArrayList<>();//burning状態の建物の集まり
		ArrayList<Building> heatingBuildings = new ArrayList<>(); //heating状態の建物の集まり
		ArrayList<Building> visibleFireBuildings = new ArrayList<>();//燃えている建物の集まり
		ArrayList<Building> visibleInfernoBuildings = new ArrayList<>();//視界内のinferno状態の建物の集まり
		ArrayList<Building> visibleBurningBuildings = new ArrayList<>();//視界内のburning状態の建物の集まり
		ArrayList<Building> visibleHeatingBuildings = new ArrayList<>(); //視界内のheating状態の建物の集まり
		List<EntityID> fireBrigadeIDs = new ArrayList<>();
		List<Integer> fireBrigadeIDsvalue = new ArrayList<>();
		int i=0;
		int j=0;
		int vertex1X;
		int vertex2X;
		int vertex1Y;
		int vertex3Y;

		for(StandardEntity entity : worldInfo.getEntitiesOfType(StandardEntityURN.BUILDING)){
			if(entity instanceof Building){
				Building building = (Building)entity;
				if(building.isOnFire()){
					fireBuildings.add(building);
				}
			}
		}

		ArrayList<FireBrigade> visibleFireBrigades = new ArrayList<>();

		for(EntityID id : worldInfo.getChanged().getChangedEntities()){
			StandardEntity standardEntity = worldInfo.getEntity(id);
			if (standardEntity instanceof FireBrigade){
				FireBrigade fireBrigade =(FireBrigade)standardEntity;
				visibleFireBrigades.add(fireBrigade);
			}
		}

		/*
		 * 視界内の建物の中で燃えている建物をvisibleFireBuildingsにいれる．
		 * また、その中でINFERNO，BURNING，HEATING状態の建物を
		 * それぞれ、visibleInfernoBuildings，visibleBurningBuildings，visibleHeatingBuildingsに入れる
		 */
		for(EntityID id : worldInfo.getChanged().getChangedEntities()){
			StandardEntity standardEntity = worldInfo.getEntity(id);
			if (standardEntity instanceof Building){
				Building building =(Building)standardEntity;
				if(building.isOnFire()){
					visibleFireBuildings.add(building);
					if(building.getFierynessEnum() == Fieryness.INFERNO){
						visibleInfernoBuildings.add(building);
					}
					else if(building.getFierynessEnum() == Fieryness.BURNING){
						visibleBurningBuildings.add(building);
					}
					else if(building.getFierynessEnum() == Fieryness.HEATING){
						visibleHeatingBuildings.add(building);
					}
				}
			}
		}
		if(!fireBuildings.isEmpty()){
			for(Building firebuilding : fireBuildings){
				if(firebuilding.getFierynessEnum() == Fieryness.INFERNO){
					infernoBuildings.add(firebuilding);
				}
				else if(firebuilding.getFierynessEnum() == Fieryness.BURNING){
					burningBuildings.add(firebuilding);
				}
				else if(firebuilding.getFierynessEnum() == Fieryness.HEATING){
					heatingBuildings.add(firebuilding);
				}
			}

			if(visibleFireBrigades.size() > 3
					|| (visibleBurningBuildings.isEmpty()&&visibleHeatingBuildings.isEmpty())
					|| !(visibleBurningBuildings.isEmpty()	&&
							visibleHeatingBuildings.isEmpty()	&&
							burningBuildings.isEmpty()			&&
							heatingBuildings.isEmpty()			&&
							!(visibleInfernoBuildings.isEmpty()))){

				//fireBuildingsの多角形を作る
				//その多角形を覆う長方形を作る
				//その長方形の各頂点の座標を取り
				//fireBuildingsの中心と頂点との距離がもっとも大きい物を半径rとする
				if(!fireBuildings.isEmpty()){
					int[] fireBuildingsx = new int[burningBuildings.size()+heatingBuildings.size()];//fireBuildingsのx座標の集まり
					int[] fireBuildingsy = new int[burningBuildings.size()+heatingBuildings.size()];//fireBuildingsのｙ座標の集まり
					for(Building building:burningBuildings){
						fireBuildingsx[i]=building.getX();
						fireBuildingsy[i]=building.getY();
						i++;
						j=i;
					}
					for(Building building:heatingBuildings){
						fireBuildingsx[i]=building.getX();
						fireBuildingsy[j]=building.getY();
						j++;
					}

					Polygon fireBuildingsPolygon = new Polygon(fireBuildingsx, fireBuildingsy, fireBuildingsx.length);//fireBuildingsの多角形
					Rectangle fireBuildingsRange = fireBuildingsPolygon.getBounds();//多角形fireBuildingsPolygonを覆う長方形
					vertex1X = fireBuildingsRange.x;//長方形fireBuildingsRangeの左上の頂点のx座標
					vertex1Y = fireBuildingsRange.y;//長方形fireBuildingsRangeの左上の頂点のy座標
					vertex2X = vertex1X + fireBuildingsRange.width;//長方形fireBuildingsRangeの右上の頂点のx座標
					vertex3Y = vertex1Y - fireBuildingsRange.height;//長方形fireBuildingsRangeの左下の頂点のy座標


					Building topBuilding = fireBuildings.get(0);
					Building bottomBuilding = fireBuildings.get(0);
					Building leftBuilding = fireBuildings.get(0);
					Building rightBuilding = fireBuildings.get(0);
					for(Building building:fireBuildings){
						if(Math.abs(building.getY()-vertex1Y) < Math.abs(topBuilding.getY()-vertex1Y)){
							topBuilding = building;
						}
						if(Math.abs(building.getY()-vertex3Y) < Math.abs(bottomBuilding.getY()-vertex3Y)){
							bottomBuilding = building;
						}
						if(Math.abs(building.getX()-vertex2X) < Math.abs(rightBuilding.getX()-vertex2X)){
							rightBuilding = building;
						}
						if(Math.abs(building.getX()-vertex1X) < Math.abs(leftBuilding.getX()-vertex1X)){
							leftBuilding = building;
						}
					}

					//ArrayList<Priority> priority = new ArrayList<>();
					ArrayList<EntityID> topNeighbor1 = new ArrayList<>();
					ArrayList<EntityID> topNeighbor2 = new ArrayList<>();

					for(EntityID entityID : topBuilding.getNeighbours()){
						if(!topNeighbor1.contains(entityID)){
							topNeighbor1.add(entityID);
						}
					}

					for(EntityID entityID : topNeighbor1){
						StandardEntity entity = this.worldInfo.getEntity(entityID);
						if(entity instanceof Road){
							Road road = (Road)entity;
							for(EntityID id : road.getNeighbours()){
								if(!topNeighbor2.contains(id)){
									topNeighbor2.add(id);
								}
							}
						}

					}


					ArrayList<EntityID> bottomNeighbor1 = new ArrayList<>();
					ArrayList<EntityID> bottomNeighbor2 = new ArrayList<>();
					ArrayList<EntityID> bottomNeighborBuildings = new ArrayList<>();

					for(EntityID entityID : bottomBuilding.getNeighbours()){
						if(!bottomNeighbor1.contains(entityID)){
							bottomNeighbor1.add(entityID);
						}
					}

					for(EntityID entityID : bottomNeighbor1){
						StandardEntity entity = this.worldInfo.getEntity(entityID);
						if(entity instanceof Road){
							Road road = (Road)entity;
							for(EntityID id : road.getNeighbours()){
								if(!bottomNeighbor2.contains(id)){
									bottomNeighbor2.add(id);
								}
							}
						}
					}



					ArrayList<EntityID> rightNeighbor1 = new ArrayList<>();
					ArrayList<EntityID> rightNeighbor2 = new ArrayList<>();
					ArrayList<EntityID> rightNeighborBuildings = new ArrayList<>();

					for(EntityID entityID : rightBuilding.getNeighbours()){
						if(!rightNeighbor1.contains(entityID)){
							rightNeighbor1.add(entityID);
						}
					}

					for(EntityID entityID :rightNeighbor1){
						StandardEntity entity = this.worldInfo.getEntity(entityID);
						if(entity instanceof Road){
							Road road = (Road)entity;
							for(EntityID id : road.getNeighbours()){
								if(!rightNeighbor2.contains(id)){
									rightNeighbor2.add(id);
								}
							}
						}
					}

					ArrayList<EntityID> leftNeighbor1 = new ArrayList<>();
					ArrayList<EntityID> leftNeighbor2 = new ArrayList<>();
					ArrayList<EntityID> leftNeighborBuildings = new ArrayList<>();

					for(EntityID entityID : leftBuilding.getNeighbours()){
						if(!leftNeighbor1.contains(entityID)){
							leftNeighbor1.add(entityID);
						}
					}

					for(EntityID entityID :leftNeighbor1){
						StandardEntity entity = this.worldInfo.getEntity(entityID);
						if(entity instanceof Road){
							Road road = (Road)entity;
							for(EntityID id : road.getNeighbours()){
								if(!leftNeighbor2.contains(id)){
									leftNeighbor2.add(id);
								}
							}
						}
					}


					for(StandardEntity entity : worldInfo.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE)){
						if(entity instanceof FireBrigade){
							FireBrigade fireBrigade =(FireBrigade) entity;
							fireBrigadeIDs.add(fireBrigade.getID());
						}
					}


					if (!fireBrigadeIDs.isEmpty()) {
						for(EntityID id : fireBrigadeIDs){
							fireBrigadeIDsvalue.add(id.getValue());
						}
						Collections.sort(fireBrigadeIDsvalue);
					}


					int number=0;
					if(!fireBrigadeIDsvalue.isEmpty()){
						for(i=0;i < fireBrigadeIDsvalue.size();i++){
							if(agentInfo.me().getID().getValue()==fireBrigadeIDsvalue.get(i)){
								if(i%4==0){
									number = 0;
								}else if(i%4==1){
									number = 1;
								}else if(i%4==2){
									number = 2;
								}else if(i%4==3){
									number = 3;
								}
							}
						}
					}
					if(number==0 && !(topNeighbor2.isEmpty())){

						//System.out.println(agentInfo.me().getID().getValue() + "getPassableMoveAction");
						return getPassableMoveAction(pathPlanning,agent.getPosition(), topNeighbor2.get(0));
					}
					if(number==1 && !(bottomNeighborBuildings.isEmpty())){

						//System.out.println(agentInfo.me().getID().getValue() + "getPassableMoveAction");
						return getPassableMoveAction(pathPlanning, agent.getPosition(), bottomNeighbor2.get(0));
					}

					if(number==2&& !(rightNeighborBuildings.isEmpty())){

						//System.out.println(agentInfo.me().getID().getValue() + "getPassableMoveAction");
						return getPassableMoveAction(pathPlanning,agent.getPosition(), rightNeighbor2.get(0));

					}

					if(number==3 && !(leftNeighborBuildings.isEmpty())){

						//System.out.println(agentInfo.me().getID().getValue() + "getPassableMoveAction");
						return getPassableMoveAction(pathPlanning, agent.getPosition(), leftNeighbor2.get(0));

					}
				}


			}
		}

		return null;


	}

	private Action calcExtinguish(FireBrigade agent, PathPlanning pathPlanning, EntityID target) {
		EntityID agentPosition = agent.getPosition();
		StandardEntity positionEntity = Objects.requireNonNull(this.worldInfo.getPosition(agent));


		List<StandardEntity> neighbourBuilding = new ArrayList<>();
		StandardEntity entity = this.worldInfo.getEntity(target);
		if(entity instanceof Building) {
			if (this.worldInfo.getDistance(positionEntity, entity) < this.maxExtinguishDistance){
				neighbourBuilding.add(entity);
			}else{
				this.result = goToNearRoad(pathPlanning, agentPosition, target);
				if(isPassable2(result)){
					return this.result;
				}else{
					//System.out.println("goToPassable");
					return goToPassableRoad2();
				}
			}
		}

		//ターゲットが視界内に入るまで移動して，放水する
		if(neighbourBuilding.size() > 0){
			neighbourBuilding.sort(new DistanceSorter(this.worldInfo, agent));

			for(EntityID id:worldInfo.getChanged().getChangedEntities()){
				StandardEntity se = worldInfo.getEntity(id);
				if(se instanceof Building && id.equals(neighbourBuilding.get(0).getID())){
					Building building = (Building)neighbourBuilding.get(0);
					if(building.isFierynessDefined() && building.isOnFire()){
						return new ActionExtinguish(neighbourBuilding.get(0).getID(), this.maxExtinguishPower);
					}else{
						this.result = goToNearRoad(pathPlanning, agentPosition, target);
						if(isPassable2(result)){

							return this.result;
						}else{
							return goToPassableRoad2();
						}
					}
				}
			}
		}
		this.result = this.getMoveAction(pathPlanning, agentPosition, target);
		if(isPassable2(result)){
			return this.result;
		}else{
			return goToPassableRoad2();
		}
	}

	private Boolean isPassable2(Action action){
		int time = agentInfo.getTime();
		FireBrigade fireBrigade = (FireBrigade)this.agentInfo.me();
		EntityID agentPosition = fireBrigade.getPosition();
		actionHistory.put(time, action);
		EntityID preAgentPosition = positionHistory.get(time-1);
		Action preAction = actionHistory.get(time-1);

		if(preAction == null || preAction == null){
			return true;
		}
		if(action != null && agentPosition != null){
			if((!action.equals(preAction) && !agentPosition.equals(preAgentPosition))){
				return true;
			}
		}

		return false;
	}

	private Action goToPassableRoad2(){

		if(agentInfo.getTime() < 3){
			return null;
		}

		int time = agentInfo.getTime();
		FireBrigade fireBrigade = (FireBrigade)this.agentInfo.me();
		EntityID agentPosition = fireBrigade.getPosition();
		EntityID prePositon = positionHistory.get(time-1);
		EntityID prePositon2 = positionHistory.get(time-2);
		StandardEntity se = worldInfo.getEntity(agentPosition);
		Action preAction = actionHistory.get(time-1);
		Action preAction2 = actionHistory.get(time-2);
		int i=0;

		if(se instanceof Road){
			Road road = (Road)se;
			List<EntityID> neighbours = road.getNeighbours();
			for(EntityID entityID : neighbours){
				i++;
				if(i>3){
					//System.out.println("break");
					break;
				}
				if(!agentPosition.equals(entityID) 
						&& !prePositon.equals(entityID) 
						&& !prePositon2.equals(entityID)){
					result = this.getMoveAction(pathPlanning, agentPosition, entityID);//TODO path生成しすぎィ！
					if(!(result.equals(preAction)) && !(result.equals(preAction2))){
						return result;
					}
				}
			}
		}

		//System.out.println("null");
		return null;
	}



	private Action getPassableMoveAction(PathPlanning pathPlanning, EntityID from, EntityID target){
		this.result = getMoveAction(pathPlanning, from, target);
		if(isPassable2(result)){

			return this.result;
		}else{

			return goToPassableRoad2();
		}
	}
	private Action getMoveAction(PathPlanning pathPlanning, EntityID from, EntityID target) {
		pathPlanning.setFrom(from);
		pathPlanning.setDestination(target);
		List<EntityID> path = pathPlanning.calc().getResult();
		if(path != null && path.size() > 0) {
			StandardEntity entity = this.worldInfo.getEntity(path.get(path.size() - 1));
			if(entity instanceof Building) {
				if(entity.getStandardURN() != StandardEntityURN.REFUGE) {
					path.remove(path.size() - 1);
				}
			}
			movetargetHistory.put(agentInfo.getTime(), target);
			return new ActionMove(path);
		}
		return null;
	}
	/*targetが燃えている建物なら、
	 * その近くの道をとって来て、
	 * その道とtargetの建物の距離が消火可能範囲内なら、
	 * その道へ移動
	 * 
	 * それ以外はtargetへ移動
	 */
	private Action goToNearRoad(PathPlanning pathPlanning, EntityID from,EntityID target){
		StandardEntity se = worldInfo.getEntity(target);
		EntityID neighbourRoad = null;
		int roadDistance;
		ArrayList<EntityID> neighbourRoads = new ArrayList<>();
		ArrayList<EntityID> neighbourRoads2 = new ArrayList<>();
		if(se instanceof Building){
			Building bu = (Building)se;
			if(bu.isOnFire()){
				for(EntityID entity:bu.getNeighbours()){
					StandardEntity st = worldInfo.getEntity(entity);
					if(st instanceof Road){
						neighbourRoad = entity;
						Road road =(Road)st;
						for(EntityID e:road.getNeighbours()){
							StandardEntity sa = worldInfo.getEntity(entity);
							if(sa instanceof Road){
								neighbourRoads.add(e);
							}
						}

					}
				}
				if(neighbourRoads.size()>0){
					for(EntityID entity:neighbourRoads){
						StandardEntity sEntity = worldInfo.getEntity(entity);
						if(sEntity instanceof Road){
							Road road =(Road)sEntity;
							for(EntityID entityID : road.getNeighbours()){
								StandardEntity se2 = worldInfo.getEntity(entity);
								if(se2 instanceof Road){
									neighbourRoads2.add(entityID);
								}
							}
						}
					}
				}
			}
		}

		if(neighbourRoads2.size()>0){
			for(EntityID entity:neighbourRoads2){
				roadDistance = worldInfo.getDistance(entity, target);
				if(roadDistance<maxExtinguishDistance){
					return this.getMoveAction(pathPlanning, from, entity);
				}
			}
		}
		if(neighbourRoads.size()>0){
			for(EntityID entity:neighbourRoads){
				roadDistance = worldInfo.getDistance(entity, target);
				if(roadDistance<maxExtinguishDistance){
					return this.getMoveAction(pathPlanning, from, entity);
				}
			}
		}else if(neighbourRoad != null){
			roadDistance = worldInfo.getDistance(neighbourRoad, target);
			if(roadDistance<maxExtinguishDistance){
				return this.getMoveAction(pathPlanning, from, neighbourRoad);
			}
		}


		return this.getMoveAction(pathPlanning, from, target);
	}

	/*
	 * rifillFlagがtrueなら，
	 * 自分の場所がREFUGEでもHYDRANTでもない，または，ある程度水がないとき
	 * Trueを返す
	 * 
	 * rifillFlagがfalseなら，水がrefillRequest以下ならtrue
	 */


}

