package RIO.extaction;

import adf.agent.action.Action;
import adf.agent.action.common.ActionMove;
import adf.agent.action.common.ActionRest;
import adf.agent.action.police.ActionClear;
import adf.agent.communication.MessageManager;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.extaction.ExtAction;
import adf.component.module.algorithm.PathPlanning;

import com.google.common.collect.Lists;

import rescuecore2.config.NoSuchConfigOptionException;
import rescuecore2.misc.Pair;
import rescuecore2.misc.geometry.GeometryTools2D;
import rescuecore2.misc.geometry.Line2D;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.misc.geometry.Vector2D;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;

import java.util.*;
import java.util.stream.Collectors;
/*
 * 作成者:林光希
 * 最終更新日:6/30
 */
public class RioneExtClear extends ExtAction {
    private PathPlanning pathPlanning;

    private int clearDistance;
    private int moveDistance;
    private int kernelTime;
    
    private HashMap<Integer, Action> actionHistory;
    private HashMap<Integer, Pair<Integer, Integer>> locationHistory;
    private HashMap<Integer, EntityID> positionHistory;
    private Set<EntityID> visitedIDs;
    
    private EntityID target;

    public RioneExtClear(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData) {
        super(ai, wi, si, moduleManager, developData);
        this.clearDistance = si.getClearRepairDistance();
        this.moveDistance = 45000; //Moveできる距離 決め打ち

        this.actionHistory = new HashMap<>();
        this.locationHistory = new HashMap<>();
        this.positionHistory = new HashMap<>();
        this.visitedIDs = new HashSet<>();
        
        this.target = null;

        switch  (si.getMode()) {
            case PRECOMPUTATION_PHASE:
                this.pathPlanning = moduleManager.getModule("ActionExtClear.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                break;
            case PRECOMPUTED:
                this.pathPlanning = moduleManager.getModule("ActionExtClear.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                break;
            case NON_PRECOMPUTE:
                this.pathPlanning = moduleManager.getModule("ActionExtClear.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                break;
        }
    }

    @Override
    public ExtAction precompute(PrecomputeData precomputeData) {
        super.precompute(precomputeData);
        if(this.getCountPrecompute() >= 2) {
            return this;
        }
        this.pathPlanning.precompute(precomputeData);
        try {
            this.kernelTime = this.scenarioInfo.getKernelTimesteps();
        }catch (NoSuchConfigOptionException e) {
            this.kernelTime = -1;
        }
        return this;
    }

    @Override
    public ExtAction resume(PrecomputeData precomputeData) {
        super.resume(precomputeData);
        if(this.getCountResume() >= 2) {
            return this;
        }
        this.pathPlanning.resume(precomputeData);
        try {
            this.kernelTime = this.scenarioInfo.getKernelTimesteps();
        }catch (NoSuchConfigOptionException e) {
            this.kernelTime = -1;
        }
        return this;
    }

    @Override
    public ExtAction preparate() {
        super.preparate();
        if(this.getCountPreparate() >= 2) {
            return this;
        }
        this.pathPlanning.preparate();
        try {
            this.kernelTime = this.scenarioInfo.getKernelTimesteps();
        }catch (NoSuchConfigOptionException e) {
            this.kernelTime = -1;
        }
        return this;
    }

    @Override
    public ExtAction updateInfo(MessageManager messageManager){
        super.updateInfo(messageManager);
        if(this.getCountUpdateInfo() >= 2) {
            return this;
        }
        this.pathPlanning.updateInfo(messageManager);
        return this;
    }

    @Override
    public ExtAction setTarget(EntityID target) {
        this.target = null;
        StandardEntity entity = this.worldInfo.getEntity(target);
        if(entity != null) {
            if (entity instanceof Road) {
                this.target = target;
            } else if (entity.getStandardURN().equals(StandardEntityURN.BLOCKADE)) {
                this.target = ((Blockade) entity).getPosition();
            } else if (entity instanceof Building) {
                this.target = target;
            }
        }
        return this;
    }

    @Override
    public ExtAction calc() {
        this.result = null;
        PoliceForce policeForce = (PoliceForce)this.agentInfo.me();
        EntityID agentPosition = policeForce.getPosition();
        StandardEntity positionEntity = Objects.requireNonNull(this.worldInfo.getEntity(agentPosition));
        
        if(this.needRest(policeForce)) {
        	if(positionEntity instanceof Refuge){
        		this.result = new ActionRest();
            	this.setHistory(policeForce, this.result);
        		return this;
        	}
        	EntityID nearestRefuge = null;
        	double minDistance = Double.MAX_VALUE;
        	for(StandardEntity refuge : this.worldInfo.getEntitiesOfType(StandardEntityURN.REFUGE)){
        		double distance = this.worldInfo.getDistance(policeForce, refuge);
        		if(distance < minDistance){
        			minDistance = distance;
        			nearestRefuge = refuge.getID();
        		}
        	}
        	this.target = nearestRefuge;
        }

        if(this.target == null) {
        	//System.out.println("target is null");
            return this;
        }
        
        StandardEntity targetEntity = this.worldInfo.getEntity(this.target);
        if(targetEntity == null || !(targetEntity instanceof Area)) {
            return this;
        }
        
        //Moveしているのに動けていない場合
        this.result = this.calcCannotMove(policeForce, positionEntity);
        if(this.result != null){
        	this.setHistory(policeForce, this.result);
        	return this;
        }
        
        //自分がRoadにいるなら(高確率でnullがreturnされる)
        if(positionEntity instanceof Road) {
            this.result = this.calcRescueAction(policeForce, (Road) positionEntity);
            if (this.result != null){
            	this.setHistory(policeForce, this.result);
                return this;
            }
        }
        
        //目的地に到達したとき
    	if(agentPosition.equals(this.target)){
    		this.result = this.calcAreaClearAction(policeForce, this.target);
    		if(this.result != null){
            	this.setHistory(policeForce, this.result);
    			return this;
    		}
    	}
        
    	//主要部分
        List<EntityID> path = this.pathPlanning.getResult(agentPosition, this.target);
        if(path == null || path.isEmpty()){
        	return this;
        }
        int index = this.getNextIndex(path, agentPosition, positionEntity);
        if(index >= 0 && index < (path.size())) {
        	this.result = this.calcNextAction(policeForce, path, index);
        	if(this.result != null){
        		this.setHistory(policeForce, this.result);
        		return this;
        	}
        }

        //Actionが決定しなかった場合、Moveする
        this.result = this.calcMoveAction(path);
        if(this.result != null){
        	this.setHistory(policeForce, this.result);
            return this;
        }
        return this;
    }
    
    //次の行動を決定する(1番よく使用するメソッド)
    private Action calcNextAction(PoliceForce policeForce, List<EntityID> path, int index){
		Pair<Integer, Integer> agentpoint = this.worldInfo.getLocation(policeForce);
		Area current = (Area)this.worldInfo.getEntity(policeForce.getPosition());
		Area next = (Area)this.worldInfo.getEntity(path.get(index));
		Area previous = this.getPreviousArea(current.getID());
        Edge edge = next.getEdgeTo(current.getID());;
        double midX = (edge.getStartX() + edge.getEndX()) / 2;
        double midY = (edge.getStartY() + edge.getEndY()) / 2;
		Vector2D clearline = new Vector2D(midX-agentpoint.first(), midY-agentpoint.second());
		Vector2D edgeline = new Vector2D(edge.getStartX()-edge.getEndX(), edge.getStartY()-edge.getEndY());
		boolean flag = (clearline.getLength() < this.clearDistance);
		
		if(this.isCurve(previous, current, next, edgeline)){
			double theta = this.getAngleTwoLines(clearline, edgeline);
			if(theta < 80){
				clearline = this.rotation(clearline, theta);
				double theta2 = this.getAngleTwoLines(clearline, edgeline);
				if(theta2 > 10){
					clearline = this.rotation(clearline, -2*theta);
				}
			}
		}
		clearline = clearline.normalised().scale(this.clearDistance * 0.8);
		
		if(current instanceof Road) {
			Road road = (Road)current;
			if(road.isBlockadesDefined() && road.getBlockades().size() > 0) {
				for(Blockade blockade : this.worldInfo.getBlockades(road)) {
					if(blockade.isApexesDefined()) {
						if(this.isClearable(policeForce, blockade, clearline)){                            	
							return new ActionClear(this.agentInfo, this.scaleClear(clearline));                               
						}
					}
				}
			}
			if(next instanceof Road && flag){
				Road road2 = (Road)next;
				if(road2.isBlockadesDefined() && road2.getBlockades().size() > 0) {
					for(Blockade blockade : this.worldInfo.getBlockades(road2)) {
						if(blockade.isApexesDefined()) {
							if(this.isClearable(policeForce, blockade, clearline)){                            	
								return new ActionClear(this.agentInfo, this.scaleClear(clearline));                               
							}
						}
					}
				}
			}
		}else if(current instanceof Building && next instanceof Road){
			Road road = (Road)next;
			if(road.isBlockadesDefined() && road.getBlockades().size() > 0) {
				for(Blockade blockade : this.worldInfo.getBlockades(road)) {
					if(blockade.isApexesDefined()) {
						if(this.isClearable(policeForce, blockade, clearline)){                            	
							return new ActionClear(this.agentInfo, this.scaleClear(clearline));                               
						}
					}
				}
			}
			return new ActionMove(Lists.newArrayList(current.getID()), (int)midX, (int)midY);
		}
		return null;
    }
    
    //road上にAT,FBがいる かつ そのAT,FBが瓦礫の中にいる ときAT,FBに向けてclearするメソッド
    private Action calcRescueAction(PoliceForce policeForce, Road road) {
		if(!road.isBlockadesDefined()){
			return null;
		}
	
		List<Human> targets = new ArrayList<>();
		for(StandardEntity se : this.worldInfo.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE,StandardEntityURN.AMBULANCE_TEAM)){
			Human human = (Human)se;
			if(!human.isPositionDefined() || human.getPosition().getValue() != road.getID().getValue()){
				continue;
			}
			if(!isInside(human)){
				continue;
			}
			targets.add(human);
		}
		if(targets.isEmpty()){
			return null;
		}
		
		Human nearestHuman = null;
		double minDistance = Double.MAX_VALUE;
		for(Human target : targets){
			double distance = this.worldInfo.getDistance(policeForce, target);
			if(distance < minDistance){
				minDistance = distance;
				nearestHuman = target;
			}
		}
		
		if(nearestHuman != null){
			if(minDistance < (this.clearDistance * 0.8)){
				Vector2D clearline = new Vector2D(nearestHuman.getX()-policeForce.getX(), nearestHuman.getY()-policeForce.getY());
				return new ActionClear(this.agentInfo,this.scaleClear(clearline));
			}else {
				return new ActionMove(Lists.newArrayList(road.getID()), nearestHuman.getX(), nearestHuman.getY());
				//return this.calcAreaClearAction(policeForce, road);
			}
		}
		return null;
	}
    
    //targetEntity上の瓦礫を消す
    private Action calcAreaClearAction(PoliceForce police, EntityID target) {
    	StandardEntity targetEntity = this.worldInfo.getEntity(target);
        if(targetEntity instanceof Building) {
            return null;
        }
        Road road = (Road)targetEntity;
        if(!road.isBlockadesDefined() || !(road.getBlockades().size() > 0)) {
            return null;
        }
        Collection<Blockade> blockades = this.worldInfo.getBlockades(road)
                .stream()
                .filter(Blockade::isApexesDefined)
                .collect(Collectors.toSet());
        
        //road上のblockadeが重なっている場合の処理(瓦礫を圧縮して消去する)
        int minDistance = Integer.MAX_VALUE;
        Blockade clearBlockade = null;
        for(Blockade blockade : blockades) {
            for(Blockade another : blockades) {
                if(!blockade.getID().equals(another.getID()) && this.intersect(blockade, another)) {
                    int distance1 = this.worldInfo.getDistance(police, blockade);
                    int distance2 = this.worldInfo.getDistance(police, another);
                    if(distance1 <= distance2 && distance1 < minDistance) {
                        minDistance = distance1;
                        clearBlockade = blockade;
                    } else if(distance2 < minDistance) {
                        minDistance = distance2;
                        clearBlockade = another;
                    }
                }
            }
        }
        if(clearBlockade != null) {
            if(minDistance < this.clearDistance*0.5) {
                return new ActionClear(clearBlockade);
            } else {
                return new ActionMove(
                        Lists.newArrayList(police.getPosition()),
                        clearBlockade.getX(),
                        clearBlockade.getY()
                );
            }
        }
        
        /*
         * 瓦礫の中で最も近い瓦礫を選択し、clearまたはmoveする
         * 瓦礫との距離は、重心との距離はなく頂点との距離で定義している(なので２重ループ)
         */
        double agentX = police.getX();
        double agentY = police.getY();
        clearBlockade = null;
        Double minPointDistance = Double.MAX_VALUE;
        int clearX = 0;
        int clearY = 0;
        for(Blockade blockade : blockades) {
            int[] apexes = blockade.getApexes();
            for (int i = 0; i < (apexes.length - 2); i += 2) {
                double distance = Math.hypot(agentX-apexes[i], agentY-apexes[i+1]);
                if(distance < minPointDistance) {
                    clearBlockade = blockade;
                    minPointDistance = distance;
                    clearX = apexes[i];
                    clearY = apexes[i + 1];
                }
            }
        }
        if(clearBlockade != null) {
        	if(minPointDistance < this.clearDistance*0.8) {
        		Vector2D clearline = new Vector2D(clearX-agentX, clearY-agentY);
        		return new ActionClear(this.agentInfo, this.scaleClear(clearline));
        	}else{
        		return new ActionMove(Lists.newArrayList(police.getPosition()), clearX, clearY);
        	}
        }
        return null;
    }
    
    //動けない場合の処理
    private Action calcCannotMove(PoliceForce policeForce, StandardEntity position) {
    	int time = this.agentInfo.getTime();
    	if(time > 5){
    		if(this.actionHistory.get(time-1) instanceof ActionMove){
    			if(this.equalsPoint(this.worldInfo.getLocation(policeForce), this.locationHistory.get(time-1))){
    				if(position instanceof Road){
    					Road road = (Road)position;
    					if(road.isBlockadesDefined() && road.getBlockades().size() > 0){
    						return this.calcAreaClearAction(policeForce, position.getID());
    					}else{
    						EntityID moveID = null;
    						for(EntityID id : road.getNeighbours()){
    							if(!this.visitedIDs.contains(id) && this.worldInfo.getEntity(id) instanceof Road){
    								moveID = id;
    								break;
    							}
    						}
    						if(moveID == null){
        						moveID = road.getNeighbours().get(0);
    						}
    						this.target = moveID;
    						return null;
    						//return new ActionMove(Lists.newArrayList(moveID));
    					}
    				}else if(position instanceof Building){
    					Building building = (Building)position;
    					Area moveArea = (Area)this.worldInfo.getEntity(building.getNeighbours().get(0));
						return new ActionMove(Lists.newArrayList(moveArea.getID()));
    				}
    			}
    		}
    	}
		return null;
	}
    
    //Moveを返す
    private ActionMove calcMoveAction(List<EntityID> path){
    	List<EntityID> movePath = Lists.newArrayList(path.get(0));
    	Pair<Integer, Integer> movePoint = this.getMovePoint(path, 0);
    	int sumDistance = 0;
    	int count = 1;
    	//エッジの中点で距離を測る版
    	while(count < path.size()){
			int distance = this.getManhattanDistance(path,count);
			if(sumDistance + distance < this.moveDistance){
				movePath.add(path.get(count));
				movePoint = this.getMovePoint(path, count);
				sumDistance += distance;
				count++;
			}else{
				if(movePath.size() == 1){
					movePath.add(path.get(count));
					movePoint = this.worldInfo.getLocation(path.get(count));
				}
				break;
			}
		}
    	return new ActionMove(movePath, movePoint.first(), movePoint.second());
    	
    	//重心で距離を測る版
    	/*while(count < path.size()){
			int distance = this.getManhattanDistance(path.get(count-1), path.get(count));
			if(sumDistance + distance < this.moveDistance){
				movePath.add(path.get(count));
				sumDistance += distance;
				count++;
			}else{
				break;
			}
		}
    	count--;
    	movePoint = this.worldInfo.getLocation(path.get(count));
    	return new ActionMove(movePath, movePoint.first(), movePoint.second());*/
	}
    
    //自分が1つ前にいたAreaを返す
    private Area getPreviousArea(EntityID currentAreaId) {
		int counter = this.agentInfo.getTime()-1;
		while(counter > 5){
			if(currentAreaId.equals(this.positionHistory.get(counter))){
				break;
			}
			counter--;
		}
		return (Area)this.worldInfo.getEntity(this.positionHistory.get(counter));
	}
    
    //p1とp2の中点を返す
    private Pair<Integer, Integer> getMiddlePoint(Point2D p1, Point2D p2) {
		int newX = (int)(p1.getX() + p2.getX())/2;
		int newY = (int)(p1.getY() + p2.getY())/2;
		return new Pair<Integer, Integer>(newX, newY);
	}
    
    //firstとsecondのマンハッタン距離を返す
    private int getManhattanDistance(EntityID first, EntityID second) {
		Pair<Integer, Integer> p1 = this.worldInfo.getLocation(first);
		Pair<Integer, Integer> p2 = this.worldInfo.getLocation(second);
		return Math.abs(p1.first()-p2.first())+Math.abs(p1.second()-p2.second());
	}
    
    private int getManhattanDistance(List<EntityID> path, int index) {
    	Area forcusArea = (Area)this.worldInfo.getEntity(path.get(index));
    	Edge edge = forcusArea.getEdgeTo(path.get(index-1));
    	Pair<Integer, Integer> center1 = getMiddlePoint(edge.getStart(), edge.getEnd());
       	Pair<Integer, Integer> center2;
    	if(index == path.size()-1){
    		center2 = this.worldInfo.getLocation(path.get(index));
    	}else{
        	Edge edge2 = forcusArea.getEdgeTo(path.get(index+1));
           	center2 = getMiddlePoint(edge2.getStart(), edge2.getEnd());
    	}
		return Math.abs(center1.first()-center2.first())+Math.abs(center1.second()-center2.second());
	}
    
    //Moveする座標を返す
    //注目エリアとその次のエリアのエッジの中点がMoveする座標
    //注目エリアがpathの最後の場合は注目エリアの重心がMoveする座標
    private Pair<Integer, Integer> getMovePoint(List<EntityID> path, int index) {
    	Area forcusArea = (Area)this.worldInfo.getEntity(path.get(index));
       	Pair<Integer, Integer> movePoint;
    	if(index == path.size()-1){
    		movePoint = this.worldInfo.getLocation(path.get(index));
    	}else{
        	Edge edge = forcusArea.getEdgeTo(path.get(index+1));
           	movePoint = getMiddlePoint(edge.getStart(), edge.getEnd());
    	}
    	return movePoint;
	}
    
    //action,location,positionを保存
    private void setHistory(PoliceForce policeForce, Action action) {
		int time = this.agentInfo.getTime();
		this.actionHistory.put(time, action);
		this.locationHistory.put(time, this.worldInfo.getLocation(policeForce));
		this.positionHistory.put(time, policeForce.getPosition());
		this.visitedIDs.add(policeForce.getPosition());
	}
    
    //p1とp2が同じ点か調べる(100mmまでの誤差は許容しているので正確にはほぼ同じかどうか)
    private boolean equalsPoint(Pair<Integer, Integer> p1, Pair<Integer, Integer> p2) {
		return Math.hypot(p1.first()-p2.first(), p1.second()-p2.second()) < 100;
	}
    
    /*
     * areaがエントランスかどうか判定するメソッド
     * エントランスの定義は隣接エリア数が3以下かつ隣接エリアにBuildingが含まれる
     * （隣接Areaに目的地になっているBuildingが含まれるかどうかを調べる）
     */
    /*private boolean isEntrance(Area area){
		List<EntityID> list = area.getNeighbours();
		if(list.size() > 3){
			return false;
		}
		for(EntityID id : list){
			StandardEntity standardEntity = this.worldInfo.getEntity(id);
			if(standardEntity instanceof Building){
				return true;
			}
		}
		return false;
	}*/
    
    //currentとnextの関係がカーブかどうかを調べる
    private boolean isCurve(Area previous, Area current, Area next, Vector2D edgeline) {
    	if(previous == null){
    		Vector2D arealine = new Vector2D(next.getX()-current.getX(), next.getY()-current.getY());
    		return (this.getAngleTwoLines(arealine, edgeline) < 70);
    	}else{
    		Vector2D v1 = new Vector2D(current.getX()-previous.getX(), current.getY()-previous.getY());
    		Vector2D v2 = new Vector2D(next.getX()-current.getX(), next.getY()-current.getY());
    		return (this.getAngleTwoLines(v1, v2) > 75);
    	}
	}
    
    //blockadeが除去可能か調べる
	private boolean isClearable(PoliceForce policeForce, Blockade blockade, Vector2D clearline) {	
		return this.intersect(policeForce.getX(), policeForce.getY(), policeForce.getX()+clearline.getX(), policeForce.getY()+clearline.getY(), blockade);
	}
    
    //2つのベクトルのなす角を鋭角で返す
    private double getAngleTwoLines(Vector2D first, Vector2D second){
		double costheta = first.dot(second) / (first.getLength()*second.getLength());
		if(costheta > 1){
			costheta = 1;
		}else if(costheta < -1){
			costheta = -1;
		}
		double theta = Math.acos(costheta)*180.0/Math.PI;
		if(theta > 90.0){
			theta = 180.0 - theta;
		}
		return theta;
	}
    
    //渡されたベクトルをtheta度だけ回転させる
    private Vector2D rotation(Vector2D vector, double theta) {
		double rad = Math.toRadians(theta);
		double sin = Math.sin(rad);
		double cos = Math.cos(rad);
		double x = vector.getX()*cos - vector.getY()*sin;
		double y = vector.getX()*sin + vector.getY()*cos;
		return new Vector2D(x, y);
	}
    
    //pathで自分がいるareaの次のarea(next)の添字を返す
    private int getNextIndex(List<EntityID> path, EntityID agentPosition, StandardEntity positionEntity) {
    	int index = path.indexOf(agentPosition);
    	if(index == -1) {
    		Area area = (Area)positionEntity;
    		for(int i = 0; i < path.size(); i++) {
    			if(area.getEdgeTo(path.get(i)) != null) {
    				index = i;
    				break;
    			}
    		}
    	}else if(index >= 0){
    		index++;
    	}
    	return index;
	}
    
    //humanが瓦礫の中にいるかどうかを返す
    private boolean isInside(Human human) {
    	StandardEntity se = this.worldInfo.getEntity(human.getPosition());
    	if(se instanceof Road){
    		Road road =(Road)se;
    		if(road.isBlockadesDefined() && road.getBlockades().size() > 0){
    			Collection<Blockade> blockades = this.worldInfo.getBlockades(road)
    					.stream()
    					.filter(Blockade::isApexesDefined)
    					.collect(Collectors.toSet());
    			for(Blockade blockade : blockades){
    				if(blockade.getShape().contains(human.getX(), human.getY())){
    					return true;
    				}
    			}
    		}
		}
		return false;
	}

    //2つのblockadeが交差するか調べる
    private boolean intersect(Blockade blockade, Blockade another) {
        if(blockade.isApexesDefined() && another.isApexesDefined()) {
            int[] apexes0 = blockade.getApexes();
            int[] apexes1 = another.getApexes();
            for (int i = 0; i < (apexes0.length - 2); i += 2) {
                for (int j = 0; j < (apexes1.length - 2); j += 2) {
                    if (java.awt.geom.Line2D.linesIntersect(
                            apexes0[i], apexes0[i + 1], apexes0[i + 2], apexes0[i + 3],
                            apexes1[j], apexes1[j + 1], apexes1[j + 2], apexes1[j + 3]
                    )) {
                        return true;
                    }
                }
            }
            for (int i = 0; i < (apexes0.length - 2); i += 2) {
                if (java.awt.geom.Line2D.linesIntersect(
                        apexes0[i], apexes0[i + 1], apexes0[i + 2], apexes0[i + 3],
                        apexes1[apexes1.length - 2], apexes1[apexes1.length - 1], apexes1[0], apexes1[1]
                )) {
                    return true;
                }
            }
            for (int j = 0; j < (apexes1.length - 2); j += 2) {
                if (java.awt.geom.Line2D.linesIntersect(
                        apexes0[apexes0.length - 2], apexes0[apexes0.length - 1], apexes0[0], apexes0[1],
                        apexes1[j], apexes1[j + 1], apexes1[j + 2], apexes1[j + 3]
                )) {
                    return true;
                }
            }
        }
        return false;
    }

    //blockadeとagentX,Y pointX,Yからなる直線が交差するか調べる
    private boolean intersect(double agentX, double agentY, double pointX, double pointY, Blockade blockade) {
        List<Line2D> lines = GeometryTools2D.pointsToLines(GeometryTools2D.vertexArrayToPoints(blockade.getApexes()), true);
        for(Line2D line : lines) {
            Point2D start = line.getOrigin();
            Point2D end = line.getEndPoint();
            double startX = start.getX();
            double startY = start.getY();
            double endX = end.getX();
            double endY = end.getY();
            if(java.awt.geom.Line2D.linesIntersect(
                    agentX, agentY, pointX, pointY,
                    startX, startY, endX, endY
            )) {
                return true;
            }
        }
        return false;
    }

    //vectorの長さをclearDistanceにする
    private Vector2D scaleClear(Vector2D vector) {
        return vector.normalised().scale(this.clearDistance);
    }

    //Restする必要があるかどうかを返す
    private boolean needRest(Human agent) {
        int hp = agent.getHP();
        int damage = agent.getDamage();
        if(damage == 0 || hp == 0) {
            return false;
        }
        int activeTime = (hp / damage) + ((hp % damage) != 0 ? 1 : 0);
        if(this.kernelTime == -1) {
            try {
                this.kernelTime = this.scenarioInfo.getKernelTimesteps();
            }catch (NoSuchConfigOptionException e) {
                this.kernelTime = -1;
            }
        }
        if(this.kernelTime == -1){
        	this.kernelTime = 300;
        }
        return (activeTime + this.agentInfo.getTime()) < this.kernelTime;
    }
    
}
