package RIO.module.complex;
//0513 calcExtinguishInside追加
//0623 クラスター廃止

import static rescuecore2.standard.entities.StandardEntityURN.AMBULANCE_TEAM;
import static rescuecore2.standard.entities.StandardEntityURN.FIRE_BRIGADE;
import static rescuecore2.standard.entities.StandardEntityURN.POLICE_FORCE;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import adf.agent.communication.MessageManager;
import adf.agent.communication.standard.bundle.MessageUtil;
import adf.agent.communication.standard.bundle.centralized.CommandFire;
import adf.agent.communication.standard.bundle.centralized.CommandPolice;
import adf.agent.communication.standard.bundle.information.MessageBuilding;
import adf.agent.communication.standard.bundle.information.MessageFireBrigade;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.communication.CommunicationMessage;
import adf.component.module.algorithm.Clustering;
import adf.component.module.complex.BuildingDetector;
import rescuecore2.misc.geometry.GeometryTools2D;
import rescuecore2.misc.geometry.Line2D;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.misc.geometry.Vector2D;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityConstants;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;

public class RioneBuildingDetector extends BuildingDetector {
	private EntityID result;

	private Clustering clustering;

	private int sendTime;

	private Collection<EntityID> agentPositions;
	private Map<EntityID, Integer> sentTimeMap;
	private int sendingAvoidTimeReceived;
	private int sendingAvoidTimeSent;
	private int commandInterval;

	EnergyFlow flow = new EnergyFlow();

	public RioneBuildingDetector(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData) {
		super(ai, wi, si, moduleManager, developData);
		switch  (si.getMode()) {
		case PRECOMPUTATION_PHASE:
			this.clustering = moduleManager.getModule("SampleBuildingDetector.Clustering", "adf.sample.module.algorithm.SampleKMeans");
			break;
		case PRECOMPUTED:
			this.clustering = moduleManager.getModule("SampleBuildingDetector.Clustering", "adf.sample.module.algorithm.SampleKMeans");
			break;
		case NON_PRECOMPUTE:
			this.clustering = moduleManager.getModule("SampleBuildingDetector.Clustering", "adf.sample.module.algorithm.SampleKMeans");
			break;
		}
		this.sendTime = 0;

		this.agentPositions = new HashSet<>();
		this.sentTimeMap = new HashMap<>();
		this.sendingAvoidTimeReceived = developData.getInteger("SampleBuildingDetector.sendingAvoidTimeReceived", 3);
		this.sendingAvoidTimeSent = developData.getInteger("SampleBuildingDetector.sendingAvoidTimeSent", 5);

	}

	@Override
	public BuildingDetector updateInfo(MessageManager messageManager) {
		super.updateInfo(messageManager);
		if(this.getCountUpdateInfo() >= 2) {
			return this;
		}
		this.clustering.updateInfo(messageManager);

		this.reflectMessage(messageManager);
		this.sendEntityInfo(messageManager);

		int currentTime = this.agentInfo.getTime();
		Human agent = (Human)this.agentInfo.me();
		int agentX = agent.getX();
		int agentY = agent.getY();
		StandardEntity positionEntity = this.worldInfo.getPosition(agent);

		Set<EntityID> changedEntities = this.worldInfo.getChanged().getChangedEntities();
		changedEntities.add(this.agentInfo.me().getID());
		//自分の視界情報を送信
		for(EntityID id : changedEntities) {
			StandardEntity standardEntity = this.worldInfo.getEntity(id);
			if(standardEntity instanceof Building){
				Building building = (Building)standardEntity;
				if(building.isOnFire()){
					messageManager.addMessage(new MessageBuilding(true, building));
				}
			}
		}

		for(CommunicationMessage message : messageManager.getReceivedMessageList()) {
			Class<? extends CommunicationMessage> messageClass = message.getClass();
			if(messageClass == MessageBuilding.class){
				MessageBuilding mb = (MessageBuilding)message;
				if(!changedEntities.contains(mb.getBuildingID())) {
					MessageUtil.reflectMessage(this.worldInfo, mb);
				}
			}else if(messageClass == MessageFireBrigade.class){
				MessageFireBrigade mfb = (MessageFireBrigade) message;
				if(!changedEntities.contains(mfb.getAgentID())){
					MessageUtil.reflectMessage(this.worldInfo, mfb);
				}
			}
		}

		if(positionEntity instanceof Road) {
			Road road = (Road)positionEntity;
			if(road.isBlockadesDefined() && road.getBlockades().size() > 0) {
				for(Blockade blockade : this.worldInfo.getBlockades(road)) {
					if(!blockade.isApexesDefined()) {
						continue;
					}
					if(this.isInside(agentX, agentY, blockade.getApexes())) {
						if ((this.sendTime + this.commandInterval) <= currentTime) {
							this.sendTime = currentTime;
							messageManager.addMessage(
									new CommandPolice(
											true,
											null,
											agent.getPosition(),
											CommandPolice.ACTION_CLEAR
											)
									);
							break;
						}
					}
				}
			}
		}
		return this;
	}

	@Override
	public BuildingDetector calc(){
		Human agent = (Human)this.agentInfo.me();
		int agentX = agent.getX();
		int agentY = agent.getY();
		StandardEntity positionEntity = this.worldInfo.getPosition(agent);
		this.result = null;


		if(positionEntity instanceof Road) {
			Road road = (Road)positionEntity;
			if(road.isBlockadesDefined() && road.getBlockades().size() > 0) {
				for(Blockade blockade : this.worldInfo.getBlockades(road)) {
					if(!blockade.isApexesDefined()) {
						continue;
					}
					if(this.isInside(agentX, agentY, blockade.getApexes())) {
						this.result = calcExtinguishInsude();
					}
				}
			}
		}
		if(this.result != null){
			return this;
		}
		
		//給水の必要性を確認する
		this.result = calcRefill();
		if(this.result != null){
			return this;
		}
		
		this.result = calcExtinguishPosition();
		if(this.result != null){
			return this;
		}

		return this;
	}
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//西田孝典
	private EntityID calcExtinguishInsude(){
		ArrayList<Priority> priority = new ArrayList<>();
		ArrayList<Building> visibleFireBuildings = new ArrayList<>();//燃えている建物の集まり
		for(EntityID id : worldInfo.getChanged().getChangedEntities()){
			StandardEntity standardEntity = worldInfo.getEntity(id);
			if (standardEntity instanceof Building){
				Building building =(Building)standardEntity;
				if(building.isOnFire()){
					visibleFireBuildings.add(building);
				}
			}
		}
		for(Building building:visibleFireBuildings){
			double value = flow.getOut(building);
			Priority priority2 = new Priority(building, value);
			priority.add(priority2);
		}
		if(!priority.isEmpty()){
			Collections.sort(priority,new PriorityComparator());
			int extinguishNumber = priority.size()/5;
			if(extinguishNumber==0){
				return priority.get(0).getBuilding().getID();
			}else{
				int targetnum = agentInfo.me().getID().getValue() % (extinguishNumber);//優先順位上位20％の建物を選ぶ
				return priority.get(targetnum).getBuilding().getID();
			}			
		}

		return null;
	}



	//インナークラス
	//西田孝典
	private class Priority{
		Building building;
		double value;

		private Priority(Building building,double value){
			this.building = building;
			this.value = value;
		}

		private Building getBuilding() {
			return this.building;
		}

		private double getValue() {
			return this.value;
		}

	}
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//西田孝典
	private EntityID calcExtinguishPosition(){
		ArrayList<Building> fireBuildings = new ArrayList<>();//燃えている建物の集まり
		ArrayList<Building> infernoBuildings = new ArrayList<>();//inferno状態の建物の集まり
		ArrayList<Building> burningBuildings = new ArrayList<>();//burning状態の建物の集まり
		ArrayList<Building> heatingBuildings = new ArrayList<>(); //heating状態の建物の集まり
		ArrayList<Building> visibleFireBuildings = new ArrayList<>();//燃えている建物の集まり
		ArrayList<Building> visibleInfernoBuildings = new ArrayList<>();//視界内のinferno状態の建物の集まり
		ArrayList<Building> visibleBurningBuildings = new ArrayList<>();//視界内のburning状態の建物の集まり
		ArrayList<Building> visibleHeatingBuildings = new ArrayList<>(); //視界内のheating状態の建物の集まり
		//ArrayList<EntityID> neighborRoads = new ArrayList<>();//建物群の近くの道の集まり
		ArrayList<Integer>  firecluster = new ArrayList<>();
		int workcluster;


		workcluster = clustering.getClusterIndex(worldInfo.getPosition(agentInfo.me().getID()));

		//担当クラスタを決める
		if(!(worldInfo.getFireBuildings().isEmpty())){
			for(Building building : worldInfo.getFireBuildings()){
				if(!firecluster.contains(clustering.getClusterIndex(building))){
					firecluster.add(clustering.getClusterIndex(building));//燃えている建物クラスタを収得
				}
			}
			workcluster = firecluster.get((agentInfo.me().getID().getValue()%firecluster.size()));
		}

		//担当クラスタ内で燃えている建物をfireBuildingsに入れる
		for(StandardEntity entity : clustering.getClusterEntities(workcluster)){
			if(entity instanceof Building){
				Building building = (Building)entity;
				if(building.isOnFire()){
					fireBuildings.add(building);
				}
			}
		}


		//fireBuildingsが空じゃなければ
		//fireBuildingsの中でINFERNO，BURNING，HEATING状態の建物を
		//それぞれ、infernoBuildings，burningBuildings，heatingBuildingsに入れる
		//fireBuildingsが空だったら、return Randomwalk();
		if(!fireBuildings.isEmpty()){
			for(Building firebuilding : fireBuildings){
				if(firebuilding.getFierynessEnum() == Fieryness.INFERNO){
					infernoBuildings.add(firebuilding);
				}
				else if(firebuilding.getFierynessEnum() == Fieryness.BURNING){
					burningBuildings.add(firebuilding);
				}
				else if(firebuilding.getFierynessEnum() == Fieryness.HEATING){
					heatingBuildings.add(firebuilding);
				}
			}


			ArrayList<FireBrigade> visibleFireBrigades = new ArrayList<>();

			for(EntityID id : worldInfo.getChanged().getChangedEntities()){
				StandardEntity standardEntity = worldInfo.getEntity(id);
				if (standardEntity instanceof FireBrigade){
					FireBrigade fireBrigade =(FireBrigade)standardEntity;
					visibleFireBrigades.add(fireBrigade);
				}
			}
			if(visibleFireBrigades.size()<4){



				/*
				 * 視界内の建物の中で燃えている建物をvisibleFireBuildingsにいれる．
				 * また、その中でINFERNO，BURNING，HEATING状態の建物を
				 * それぞれ、visibleInfernoBuildings，visibleBurningBuildings，visibleHeatingBuildingsに入れる
				 */
				ArrayList<Priority> priority = new ArrayList<>();
				for(EntityID id : worldInfo.getChanged().getChangedEntities()){
					StandardEntity standardEntity = worldInfo.getEntity(id);
					if (standardEntity instanceof Building){
						Building building =(Building)standardEntity;
						if(building.isOnFire()){
							visibleFireBuildings.add(building);
							if(building.getFierynessEnum() == Fieryness.INFERNO){
								visibleInfernoBuildings.add(building);
							}
							else if(building.getFierynessEnum() == Fieryness.BURNING){
								visibleBurningBuildings.add(building);
							}
							else if(building.getFierynessEnum() == Fieryness.HEATING){
								visibleHeatingBuildings.add(building);
							}
						}
					}
				}

				/*
				 * visibleBurningBuildingsまたはvisibleHeatingBuildingsが空でないとき
				 * visibleBurningBuildingsとvisibleHeatingBuildingsすべての評価値をとって
				 * 優先順位上位20％の建物を消火
				 */
				if(!(visibleBurningBuildings.isEmpty()&&visibleHeatingBuildings.isEmpty())){

					for(Building building:visibleHeatingBuildings){
						double value = flow.getOut(building);
						Priority priority2 = new Priority(building, value);
						priority.add(priority2);
					}

					for(Building building:visibleBurningBuildings){
						double value = flow.getOut(building);
						Priority priority2 = new Priority(building, value);
						priority.add(priority2);
					}
					if(!priority.isEmpty()){
						Collections.sort(priority,new PriorityComparator());
						int extinguishNumber = priority.size()/5;
						if(extinguishNumber==0){
							return priority.get(0).getBuilding().getID();
						}else{
							int targetnum = agentInfo.me().getID().getValue() % (extinguishNumber);//優先順位上位20％の建物を選ぶ
							return priority.get(targetnum).getBuilding().getID();
						}			
					}
				}

				if(visibleBurningBuildings.isEmpty()		&&
						visibleHeatingBuildings.isEmpty()	&&
						burningBuildings.isEmpty()			&&
						heatingBuildings.isEmpty()			&&
						!(visibleInfernoBuildings.isEmpty())){

					for(Building building:visibleInfernoBuildings){
						double value = flow.getOut(building);
						Priority priority2 = new Priority(building, value);
						priority.add(priority2);
					}
					if(!priority.isEmpty()){
						Collections.sort(priority,new PriorityComparator());
						int extinguishNumber = priority.size()/5;
						if(extinguishNumber==0){
							return priority.get(0).getBuilding().getID();
						}else{
							int targetnum = agentInfo.me().getID().getValue() % (extinguishNumber);//優先順位上位20％の建物を選ぶ
							return priority.get(targetnum).getBuilding().getID();
						}			
					}
				}
			}


		}
		//this.result = randomWalk3();
		this.result = null;
		return this.result;
	}


	//西田孝典


	//森本理沙，中村仁

	//中村仁
	private class FlowLine {
		private final Building building;
		private final double distance; 
		private final double sin;
		public FlowLine(Building target, Collection<StandardEntity> range, Line2D ray) {
			// 線分rayの中で最も近くに当たっているBuildingを求める
			Building nearestHit = null;
			Line2D hitLine = null;
			double nearestHitDist = Double.MAX_VALUE;

			for (StandardEntity se : range) {
				if (!(se instanceof Building) || target.equals(se))
					continue;
				final Building building = (Building) se;
				for (Edge e : building.getEdges()) {
					if (e.isPassable()) continue;
					final Line2D l = e.getLine();
					final double d1 = ray.getIntersection(l);
					final double d2 = l.getIntersection(ray);
					if (d1 < nearestHitDist
							&& 0.0 <= d2 && d2 <= 1.0
							&& 0.0 < d1 && d1 <= 1.0) {
						nearestHit = building;
						hitLine = l;
						nearestHitDist = d1;
					}
				}
			}
			building = nearestHit;
			if (nearestHit != null) {
				final Point2D rayPt1 = ray.getOrigin(), rayPt2 = ray.getEndPoint();
				final Point2D hitLinePt1 = hitLine.getOrigin(), hitLinePt2 = hitLine.getEndPoint();
				// 外積
				final double cross = Math.abs(
						(rayPt2.getX() - rayPt1.getX()) * (hitLinePt2.getY() - hitLinePt1.getY())
						- (hitLinePt2.getX() - hitLinePt1.getX()) * (rayPt2.getY() - rayPt1.getY()));
				final double rayLength = GeometryTools2D.getDistance(rayPt1, rayPt2);
				final double hitLineLength = GeometryTools2D.getDistance(hitLinePt1, hitLinePt2);
				sin = cross / (rayLength * hitLineLength);
				// 交点までのhitLineと垂直な方向の距離
				distance = rayLength * nearestHitDist;
			}
			else {
				sin = 0.0;
				distance = 0.0;
			}
		}

		public Building getBuilding() {
			return building;
		}

		public double getDistance() {
			return distance;
		}

		public double getSin() {
			return sin;
		}
	}

	public class EnergyFlow {

		/** 放射状の線は sqrt(建物の面積) * LINE_LENGTH_SCALE となる */
		//private DynamicConstant LINE_LENGTH_SCALE = DynamicConstant.create("1000.0 * 15.0");
		private static final double LINE_LENGTH_SCALE = 1000.0 * 2.0;
		/** 単位調整定数 */
		private static final double K = 1000000.0;// = k / (4 * PI)
		private static final int rayCount = 32;

		private final Map<Building, FlowLine[]> flowGraph;

		/** そのBuildingが与える影響． (温度 / 交点との距離の2乗) の合計 */
		private final Map<Building, Double> outFlow;
		/** そのBuildingが受ける影響．*/
		private final Map<Building, Double> inFlow;
		/** そのBuildingが受けた影響の合計．*/
		private final Map<Building, Double> inTotal;

		private ArrayList<Building> affectedRanking;

		public EnergyFlow() {

			//Buildingの塊取ってくる
			final Collection<StandardEntity> buildings = worldInfo.getEntitiesOfType(StandardEntityURN.BUILDING);

			// rayCount本の放射状の線分
			Map<Building, FlowLine[]> modifiableFlowGraph = new HashMap<Building, FlowLine[]>(buildings.size());

			for (StandardEntity se : buildings) {
				final Building target = (Building) se;
				//TODO isGroundAreaDefend
				final double r = Math.sqrt(target.getGroundArea()) * LINE_LENGTH_SCALE;
				FlowLine[] lineArray = new FlowLine[rayCount];
				final double dAngle = Math.PI * 2 / rayCount;
				for (int i = 0; i < rayCount; ++i) {
					final double angle = i * dAngle;
					final double dx = Math.sin(angle) * r;
					final double dy = Math.cos(angle) * r;
					final Line2D ray = new Line2D(target.getX(), target.getY(), dx, dy);
					lineArray[i] = new FlowLine(target, worldInfo.getObjectsInRange(target, (int) r), ray);
				}
				modifiableFlowGraph.put(target, lineArray);
			}
			flowGraph = Collections.unmodifiableMap(modifiableFlowGraph);


			outFlow = new HashMap<Building, Double>(buildings.size());
			inFlow = new HashMap<Building, Double>(buildings.size());
			inTotal = new HashMap<Building, Double>(buildings.size());
		}

		public void update(ChangeSet change) {
			inFlow.clear();
			// 与える影響を更新
			for (Building building : worldInfo.getFireBuildings()) {
				// calcurateLineHits内部で受ける影響も更新される
				final double lineAffecct = calcurateLineHits(building);
				outFlow.put(building, lineAffecct);
			}
			////System.out.println(affectedFlow.size());
			// 燃えていなければ消す
			for (Iterator<Building> it = outFlow.keySet().iterator(); it.hasNext();) {
				Building building = it.next();
				if (worldInfo.getFireBuildings().contains(building))
					continue;
				if (!building.isOnFire()) {
					it.remove();
				}
			}
			// 各建物合計を計算
			affectedRanking = new ArrayList<Building>();
			for (Entry<Building, Double> entry : inFlow.entrySet()) {
				Building key = entry.getKey();
				if (key.isFierynessDefined()
						&& key.getFierynessEnum() == StandardEntityConstants.Fieryness.BURNT_OUT) {
					inTotal.remove(key);
					continue;
				}
				inTotal.put(key, getInTotal(key) + entry.getValue());
				affectedRanking.add(key);
			}
			// 降順ソート
			Collections.sort(affectedRanking, new Comparator<Building>() {
				@Override
				public int compare(Building b1, Building b2) {
					double d = getInTotal(b1) - getInTotal(b2);
					if (d < 0) return -1;
					if (d > 0) return 1;
					return 0;
				}
			});
		}

		private double calcurateLineHits(Building target) {
			if (!target.isTemperatureDefined()) return 0.0;
			FlowLine[] lines = flowGraph.get(target);
			if (lines == null) return 0.0;

			double result = 0.0;
			for (FlowLine line : lines) {
				final Building nearestHit = line.getBuilding();
				if (nearestHit == null) continue;
				// 燃え尽きていたら影響なしとして考える
				if (nearestHit.isFierynessDefined()
						&& nearestHit.getFierynessEnum() == StandardEntityConstants.Fieryness.BURNT_OUT)
					continue;
				final double dist = line.getDistance();
				final double sin = line.getSin();
				final double affectTemp = target.getTemperature();
				final double value = K * sin * affectTemp / (dist * dist);

				Double preAffectedValue = inFlow.get(nearestHit);
				if (preAffectedValue == null) {
					preAffectedValue = 0.0;
				}
				inFlow.put(nearestHit, preAffectedValue + value);

				result += value;
				// まだ燃えたことかがないBuildingに与える影響は2倍にする
				if (nearestHit.isFierynessDefined()
						&& nearestHit.getFierynessEnum() == Fieryness.UNBURNT) {
					result += value;
				}
			}
			return result;
		}

		/**
		 * 建物の与える影響を取得します．
		 * この値は周りの建物の距離と温度によります．
		 * @param building 対象の建物
		 * @return buildingの与える影響
		 */
		public double getOut(Building building) {
			Double result = outFlow.get(building);
			if (result == null) return 0.0;
			return result;
		}
		/**
		 * 建物の与える影響を取得します．
		 * この値は周りの建物の距離と温度によります．
		 * @param id 対象の建物
		 * @return buildingの与える影響
		 */
		public double getOut(EntityID id) {
			return getOut((Building) worldInfo.getEntity(id));
		}

		/**
		 * 建物の受ける影響を取得します．
		 * この値は周りの建物の距離と温度によります．
		 * @param building 対象の建物
		 * @return buildingの受ける影響
		 */
		public double getIn(Building building) {
			Double result = inFlow.get(building);
			if (result == null) return 0.0;
			return result;
		}
		/**
		 * 建物の受ける影響を取得します．
		 * この値は周りの建物の距離と温度によります．
		 * @param id 対象の建物
		 * @return buildingの受ける影響
		 */
		public double getIn(EntityID id) {
			return getIn((Building) worldInfo.getEntity(id));
		}

		/**
		 * 建物の受けた影響の合計を取得します．
		 * この値は周りの建物の距離と温度によります．
		 * @param building 対象の建物
		 * @return buildingの受けた影響
		 */
		public double getInTotal(Building building) {
			Double result = inTotal.get(building);
			if (result == null) return 0.0;
			return result;
		}
		/**
		 * 建物の受けた影響の合計を取得します．
		 * この値は周りの建物の距離と温度によります．
		 * @param id 対象の建物
		 * @return buildingの受けた影響
		 */
		public double getInTotal(EntityID id) {
			return getInTotal((Building) worldInfo.getEntity(id));
		}

		/**
		 * 影響を受ける順で並べられたリストを返します．
		 * 先頭であるほど燃えている可能性が高いです．
		 * @return 影響を受ける順ので並べられたリスト．ただしこのリストは読み取り専用です．
		 * 必ずしもBuildingが燃えてないことに注意してください．
		 */
		public List<Building> getInTotalRanking() {
			return Collections.unmodifiableList(affectedRanking);
		}
	}
	//中村仁
	class PriorityComparator implements Comparator<Priority> {

		//比較メソッド（データクラスを比較して-1, 0, 1を返すように記述する）
		public int compare(Priority a, Priority b) {
			double no1 = a.getValue();
			double no2 = b.getValue();

			if (no1 > no2) {
				return -1;
			} else if (no1 == no2) {
				return 0;

			} else {
				return 1;

			}
		}
	}



	@Override
	public EntityID getTarget() {
		return this.result;
	}

	@Override
	public BuildingDetector precompute(PrecomputeData precomputeData) {
		super.precompute(precomputeData);
		if(this.getCountPrecompute() >= 2) {
			return this;
		}
		this.clustering.precompute(precomputeData);
		return this;
	}

	@Override
	public BuildingDetector resume(PrecomputeData precomputeData) {
		super.resume(precomputeData);
		if(this.getCountPrecompute() >= 2) {
			return this;
		}
		this.clustering.resume(precomputeData);
		return this;
	}

	@Override
	public BuildingDetector preparate() {
		super.preparate();
		if(this.getCountPrecompute() >= 2) {
			return this;
		}
		this.clustering.preparate();
		return this;
	}

	private boolean isInside(double pX, double pY, int[] apex) {
		Point2D p = new Point2D(pX, pY);
		Vector2D v1 = (new Point2D(apex[apex.length - 2], apex[apex.length - 1])).minus(p);
		Vector2D v2 = (new Point2D(apex[0], apex[1])).minus(p);
		double theta = this.getAngle(v1, v2);

		StandardEntity se = agentInfo.me();
		Human me = (Human) se;

		if(me.getBuriedness() == 0) {
			return false;
		}

		for(int i = 0; i < apex.length - 2; i += 2) {
			v1 = (new Point2D(apex[i], apex[i + 1])).minus(p);
			v2 = (new Point2D(apex[i + 2], apex[i + 3])).minus(p);
			theta += this.getAngle(v1, v2);
		}
		return Math.round(Math.abs((theta / 2) / Math.PI)) >= 1;
	}

	private double getAngle(Vector2D v1, Vector2D v2) {
		double flag = (v1.getX() * v2.getY()) - (v1.getY() * v2.getX());
		double angle = Math.acos(((v1.getX() * v2.getX()) + (v1.getY() * v2.getY())) / (v1.getLength() * v2.getLength()));
		if(flag > 0) {
			return angle;
		}
		if(flag < 0) {
			return -1 * angle;
		}
		return 0.0D;
	}


	private void reflectMessage(MessageManager messageManager) {
		Set<EntityID> changedEntities = this.worldInfo.getChanged().getChangedEntities();
		changedEntities.add(this.agentInfo.getID());
		int time = this.agentInfo.getTime();
		for(CommunicationMessage message : messageManager.getReceivedMessageList(MessageBuilding.class)) {
			MessageBuilding mb = (MessageBuilding)message;
			if(!changedEntities.contains(mb.getBuildingID())) {
				MessageUtil.reflectMessage(this.worldInfo, mb);
			}
			this.sentTimeMap.put(mb.getBuildingID(), time + this.sendingAvoidTimeReceived);
		}
	}

	private boolean checkSendFlags(){
		boolean isSendBuildingMessage = true;

		StandardEntity me = this.agentInfo.me();
		if(!(me instanceof Human)){
			return false;
		}
		Human agent = (Human)me;
		EntityID agentID = agent.getID();
		EntityID position = agent.getPosition();
		StandardEntityURN agentURN = agent.getStandardURN();
		EnumSet<StandardEntityURN> agentTypes = EnumSet.of(AMBULANCE_TEAM, FIRE_BRIGADE, POLICE_FORCE);
		agentTypes.remove(agentURN);

		this.agentPositions.clear();
		for(StandardEntity entity : this.worldInfo.getEntitiesOfType(agentURN)) {
			Human other = (Human)entity;
			if(isSendBuildingMessage) {
				if (other.getPosition().getValue() == position.getValue()) {
					if (other.getID().getValue() > agentID.getValue()) {
						isSendBuildingMessage = false;
					}
				}
			}
			this.agentPositions.add(other.getPosition());
		}
		for(StandardEntityURN urn : agentTypes) {
			for(StandardEntity entity : this.worldInfo.getEntitiesOfType(urn)) {
				Human other = (Human) entity;
				if(isSendBuildingMessage) {
					if (other.getPosition().getValue() == position.getValue()) {
						if (urn == FIRE_BRIGADE) {
							isSendBuildingMessage = false;
						} else if (agentURN != FIRE_BRIGADE && other.getID().getValue() > agentID.getValue()) {
							isSendBuildingMessage = false;
						}
					}
				}
				this.agentPositions.add(other.getPosition());
			}
		}
		return isSendBuildingMessage;
	}

	private void sendEntityInfo(MessageManager messageManager) {
		if(this.checkSendFlags()) {
			Building building = null;
			int currentTime = this.agentInfo.getTime();
			Human agent = (Human) this.agentInfo.me();
			for (EntityID id : this.worldInfo.getChanged().getChangedEntities()) {
				StandardEntity entity = this.worldInfo.getEntity(id);
				if (entity instanceof Building) {
					Integer time = this.sentTimeMap.get(id);
					if (time != null && time > currentTime) {
						continue;
					}
					Building target = (Building) entity;
					if (!this.agentPositions.contains(target.getID())) {
						building = this.selectBuilding(building, target);
					} else if (target.getID().getValue() == agent.getPosition().getValue()) {
						building = this.selectBuilding(building, target);
					}
				}
			}
			if (building != null) {
				messageManager.addMessage(new MessageBuilding(true, building));
				this.sentTimeMap.put(building.getID(), currentTime + this.sendingAvoidTimeSent);
			}
		}
	}

	private Building selectBuilding(Building building1, Building building2) {
		if(building1 != null) {
			if(building2 != null) {
				if(building1.isOnFire() && building2.isOnFire()) {
					if (building1.getFieryness() < building2.getFieryness()) {
						return building2;
					} else if (building1.getFieryness() > building2.getFieryness()) {
						return building1;
					}
					if(building1.isTemperatureDefined() && building2.isTemperatureDefined()) {
						return building1.getTemperature() < building2.getTemperature() ? building2 : building1;
					}
				} else if (!building1.isOnFire() && building2.isOnFire()) {
					return building2;
				}
			}
			return building1;
		}
		return building2 != null ? building2 : null;
	}


	private EntityID calcRefill(){
		FireBrigade fireBrigade = (FireBrigade)this.agentInfo.me();
		EntityID agentPosition = fireBrigade.getPosition();
		int water = fireBrigade.getWater();//水の量
		StandardEntityURN positionURN = this.worldInfo.getPosition(fireBrigade).getStandardURN();
		int maxWater = scenarioInfo.getFireTankMaximum();
		int Damage = fireBrigade.getDamage();
		
		//ポジションがRefugeなら9割いくまで補給する
		if(positionURN.equals(StandardEntityURN.REFUGE)){
			if (water < maxWater*0.9){
				return agentPosition;
			}else{ //9割いってたら給水やめる
				return null;
			}
		}
		//ダメージが0より上なら，リフジに強制移動
		if(Damage > 0){
			for(StandardEntity se: worldInfo.getEntitiesOfType(StandardEntityURN.REFUGE)) {
				return se.getID();
			}
		}

		//1割切ったら給水しにいく
		if(water < maxWater*0.1) {
			for(StandardEntity se: worldInfo.getEntitiesOfType(StandardEntityURN.REFUGE)) {
				return se.getID();
			}
		}

		
		return null;

	}

}

