package RIO.module.complex;

import static rescuecore2.standard.entities.StandardEntityURN.AMBULANCE_TEAM;
import static rescuecore2.standard.entities.StandardEntityURN.BLOCKADE;
import static rescuecore2.standard.entities.StandardEntityURN.BUILDING;
import static rescuecore2.standard.entities.StandardEntityURN.CIVILIAN;
import static rescuecore2.standard.entities.StandardEntityURN.FIRE_BRIGADE;
import static rescuecore2.standard.entities.StandardEntityURN.HYDRANT;
import static rescuecore2.standard.entities.StandardEntityURN.POLICE_FORCE;
import static rescuecore2.standard.entities.StandardEntityURN.REFUGE;
import static rescuecore2.standard.entities.StandardEntityURN.ROAD;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import adf.agent.communication.MessageManager;
import adf.agent.communication.standard.bundle.MessageUtil;
import adf.agent.communication.standard.bundle.centralized.CommandPolice;
import adf.agent.communication.standard.bundle.information.MessageAmbulanceTeam;
import adf.agent.communication.standard.bundle.information.MessageBuilding;
import adf.agent.communication.standard.bundle.information.MessageCivilian;
import adf.agent.communication.standard.bundle.information.MessageFireBrigade;
import adf.agent.communication.standard.bundle.information.MessagePoliceForce;
import adf.agent.communication.standard.bundle.information.MessageRoad;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.communication.CommunicationMessage;
import adf.component.module.algorithm.Clustering;
import adf.component.module.algorithm.PathPlanning;
import adf.component.module.complex.RoadDetector;
import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.EntityID;

public class RioneRoadDetector extends RoadDetector {
    private Map<EntityID, Double> targetPriorities;
    private ArrayList<Human> buriedHumans;

    private PathPlanning pathPlanning;
    private Clustering clustering;
    
    private Set<EntityID> visitedIds; //林追加部分

    private EntityID result;

    public RioneRoadDetector(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData) {
        super(ai, wi, si, moduleManager, developData);
        switch  (scenarioInfo.getMode()) {
            case PRECOMPUTATION_PHASE:
                this.pathPlanning = moduleManager.getModule("SampleRoadDetector.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                this.clustering = moduleManager.getModule("SampleBuildingDetector.Clustering", "adf.sample.module.algorithm.SampleKMeans");
                break;
            case PRECOMPUTED:
                this.pathPlanning = moduleManager.getModule("SampleRoadDetector.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                this.clustering = moduleManager.getModule("SampleBuildingDetector.Clustering", "adf.sample.module.algorithm.SampleKMeans");
                break;
            case NON_PRECOMPUTE:
                this.pathPlanning = moduleManager.getModule("SampleRoadDetector.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                this.clustering = moduleManager.getModule("SampleBuildingDetector.Clustering", "adf.sample.module.algorithm.SampleKMeans");
                break;
        }
        this.result = null;
        this.buriedHumans = new ArrayList<>();
        this.visitedIds = new HashSet<>(); //林追加部分
    }

    @Override
    public RoadDetector calc() {
        if(this.result == null) {
        	//System.out.println(agentInfo.getTime()+" change target "+agentInfo.getID());
            EntityID positionID = this.agentInfo.getPosition();
            if(this.targetPriorities.containsKey(positionID) && this.targetPriorities.get(positionID) > 0){
                this.result = positionID;
                return this;
            }
            if(this.agentInfo.getTime() < 5 && this.clustering.getClusterIndex(this.agentInfo.me().getID()) != this.clustering.getClusterIndex(this.agentInfo.getPosition())){
            	resultToCluster();
            }else{ // agentInfo.getTime() >= 5
            	selectTop();
            }
            //System.out.println(this.result+" "+targetPriorities.get(this.result));
            this.visitedIds.add(this.result); //林追加部分
        }
        return this;
    }

    private void resultToCluster() {
    	int index = this.clustering.getClusterIndex(this.agentInfo.me().getID());
    	this.clustering.getClusterEntities(index).stream() //林追加部分
    		.filter(se -> se instanceof Refuge)
    		.map(se -> ((Area)se).getNeighbours())
		.	forEach(set -> {
    			set.stream()
    				.filter(ne -> worldInfo.getEntity(ne) instanceof Road)
    				.forEach(id -> this.result = id);
			});
    	if(this.result == null){
    		this.clustering.getClusterEntities(index).stream()
    			.filter(se -> se instanceof Road)
    			.map(se -> se.getID())
    			.limit(1)
    			.forEach(id -> this.result = id);
    	}
	}
    
    private void selectTop() {
    	targetPriorities.entrySet().stream()
	        .sorted(java.util.Collections.reverseOrder(java.util.Map.Entry.comparingByValue()))
	        .limit(1)
	        .forEach(e -> {
	        	//System.out.println(e.getKey() + " : " + e.getValue());
	        	this.result = e.getKey();
	        });
	}
    
    @Override
    public EntityID getTarget() {
        return this.result;
    }

    @Override
    public RoadDetector precompute(PrecomputeData precomputeData) {
        super.precompute(precomputeData);
        if(this.getCountPrecompute() >= 2) {
            return this;
        }
        this.pathPlanning.precompute(precomputeData);
        this.clustering.precompute(precomputeData);
        return this;
    }

    @Override
    public RoadDetector resume(PrecomputeData precomputeData) {
        super.resume(precomputeData);
        if(this.getCountResume() >= 2) {
            return this;
        }
        this.pathPlanning.resume(precomputeData);
        this.clustering.resume(precomputeData);
        this.targetPriorities = new Hashtable<>();
        return this;
    }

    @Override
    public RoadDetector preparate() {
        super.preparate();
        if(this.getCountPreparate() >= 2) {
            return this;
        }
        this.pathPlanning.preparate();
        this.clustering.preparate();
        this.targetPriorities = new Hashtable<>();
        return this;
    }
    
    @Override
    public RoadDetector updateInfo(MessageManager messageManager) {
    	
    	this.clustering.updateInfo(messageManager);
    	ArrayList<Human> activeHumans = new ArrayList<>();
    	
    	//視界情報の更新
        Set<EntityID> changedEntities = this.worldInfo.getChanged().getChangedEntities();
        changedEntities.add(this.agentInfo.me().getID());
        //自分の視界情報を送信
        for(EntityID id : changedEntities) {
        	StandardEntity standardEntity = this.worldInfo.getEntity(id);
        	if(standardEntity instanceof Civilian) {
        		Civilian civilian = (Civilian)standardEntity;
        		StandardEntity position = this.worldInfo.getEntity(civilian.getPosition());
        		StandardEntityURN positionURN = position.getStandardURN();
        		if(positionURN != REFUGE){
        			if(civilian.isHPDefined() && civilian.getHP() > 0){
        				if((civilian.isBuriednessDefined() && civilian.getBuriedness() > 0) ||
        						(civilian.isDamageDefined() && civilian.getDamage() > 0)){
                			messageManager.addMessage(new MessageCivilian(true, civilian));
        				}
        			}
        		}
        	}else if(standardEntity instanceof Building){
        		Building building = (Building)standardEntity;
        		messageManager.addMessage(new MessageBuilding(true, building));
        	}
        }    	
    	
    	//他Agentが送信した情報を受信して処理
        for(CommunicationMessage message : messageManager.getReceivedMessageList()) {
        	Class<? extends CommunicationMessage> messageClass = message.getClass();
        	if(messageClass == MessageCivilian.class) {
        		MessageCivilian mc = (MessageCivilian) message;
        		if(!changedEntities.contains(mc.getAgentID())){
        			MessageUtil.reflectMessage(this.worldInfo, mc);
        		}
        	} else if(messageClass == MessageBuilding.class){
        		MessageBuilding mb = (MessageBuilding)message;
        		if(!changedEntities.contains(mb.getBuildingID())) {
        			MessageUtil.reflectMessage(this.worldInfo, mb);
        		}
        	} else if(messageClass == MessageAmbulanceTeam.class) {
        		MessageAmbulanceTeam mat = (MessageAmbulanceTeam)message;
        		if(!changedEntities.contains(mat.getAgentID())) {
        			MessageUtil.reflectMessage(this.worldInfo, mat);
        			prioritizeByMessage(mat);
        		}
        		if(mat.getAction() == MessageAmbulanceTeam.ACTION_RESCUE || mat.getAction() == MessageAmbulanceTeam.ACTION_LOAD){
        			if(mat.getAgentID() != this.agentInfo.me().getID()){
        				activeHumans.add((Human)this.worldInfo.getEntity(mat.getTargetID()));
        			}
        		}
        	} else if(messageClass == MessageFireBrigade.class) {
        		MessageFireBrigade mfb = (MessageFireBrigade) message;
        		if(!changedEntities.contains(mfb.getAgentID())) {
        			MessageUtil.reflectMessage(this.worldInfo, mfb);
        			prioritizeByMessage(mfb);
        		}
        	} else if(messageClass == MessagePoliceForce.class) {
        		MessagePoliceForce mpf = (MessagePoliceForce) message;
        		if(!changedEntities.contains(mpf.getAgentID())) {
        			MessageUtil.reflectMessage(this.worldInfo, mpf);
        		    prioritizeByMessage(mpf);
        		}
        	} else if(messageClass == CommandPolice.class) {
                prioritizeByMessage((CommandPolice)message);
            } else if(messageClass == MessageRoad.class) {
                prioritizeByMessage((MessageRoad)message, changedEntities);
            }
        }
    	  	
        this.buriedHumans.clear();
        int clusterIndex = clustering.getClusterIndex(this.agentInfo.me().getID());
        Collection<StandardEntity> elements = clustering.getClusterEntities(clusterIndex);
        
        for (StandardEntity next : this.worldInfo.getEntitiesOfType(CIVILIAN,FIRE_BRIGADE, POLICE_FORCE)) {
        	Human h = (Human) next;
        	StandardEntity positionEntity = this.worldInfo.getPosition(h);
        	StandardEntityURN positionURN = positionEntity.getStandardURN();
        	if(positionEntity != null && positionEntity instanceof Building) {
        		if (elements.contains(positionEntity) || elements.contains(h)) {
        			if (h.isHPDefined() && h.getHP() > 0) {
        				if ((h.isBuriednessDefined() && h.getBuriedness() > 0) ||
        						(h.isDamageDefined() && h.getDamage() > 0) &&
        						(positionURN != REFUGE) && positionURN != AMBULANCE_TEAM) {
        					this.buriedHumans.add(h);
        				}
        			}
        		}
        	}
        }
        this.buriedHumans.removeAll(activeHumans);
    	
    	// キーにroadID，値に優先度をもったMapを優先度0で初期化する    	
    	initPriorities();
    	// 瓦礫に挟まれているエージェントのいるroadの優先度を上げる
    	prioritizeByAgents();
    	//交差点の優先度を上げる
    	prioritizeByIntersections();
    	// 自分のクラスター内の優先度を上げる
    	prioritizeByCluster();
    	// refugeのエントランスの優先度を上げる
    	prioritizeByRefuge();
    	//建物の中にいる人数が多ければ優先度を上げる
    	prioritizeByCrowdMember();
    	// 燃え尽きた建物の半径10000mmのroadの優先度を下げる
    	deprioritizeByBuilding();
    	// 瓦礫が無いroadの優先度を-1にする
    	deprioritizeByBlockade();
    	//林追加部分
    	deprioritizeByVisited();
        super.updateInfo(messageManager);
        if(this.getCountUpdateInfo() >= 2) {
            return this;
        }
        this.pathPlanning.updateInfo(messageManager);
        if(this.result != null) {
        	if(this.agentInfo.getPosition().equals(this.result)) {
                StandardEntity entity = this.worldInfo.getEntity(this.result);
                if(entity instanceof Building) {
                    this.result = null;
                }else if(entity instanceof Road) {
                    Road road = (Road)entity;
                    if(!road.isBlockadesDefined() || road.getBlockades().isEmpty()) {
                    	//System.out.println("No blockades");
                    	this.result = null;
                    }
                }
            }
        }
        /*
    	System.out.println("■" + agentInfo.me().getID());
        targetPriorities.entrySet().stream()
        	.filter(e -> e.getValue() > 0)
	        .sorted(java.util.Collections.reverseOrder(java.util.Map.Entry.comparingByValue()))
	        .limit(5)
	        .forEach(e -> {
	        	System.out.println("Road = " + e.getKey() + " : " + e.getValue());
	        });
	    */
        return this;
    }

    private void prioritizeByAgents() {
    	for(StandardEntity agent : this.worldInfo.getEntitiesOfType(FIRE_BRIGADE, AMBULANCE_TEAM)){
    		Human human = (Human)agent;
    		StandardEntity positionEntity = this.worldInfo.getPosition(human);
    		if(positionEntity instanceof Road){
    		    Road road = (Road)positionEntity;
    		    if(road.isBlockadesDefined() && road.getBlockades().size() > 0){
    		    	for(Blockade blockade : this.worldInfo.getBlockades(road)){
    		            if(!blockade.isApexesDefined()){
    		                continue;
    		            }
    		            if(this.isInBlockade(human)){
    		            	EntityID keyID = road.getID();
    		    			if(targetPriorities.containsKey(keyID)){
    			    			double tempValue = targetPriorities.get(keyID);
    			    			if(agent instanceof AmbulanceTeam){
    			    				targetPriorities.put(keyID, tempValue+1);
    			    				//System.out.println("AT In Blockade ID = " + agent.getID() + " Road ID = " + keyID);
    			    			}
    			    			if(agent instanceof FireBrigade){
    			    				targetPriorities.put(keyID, tempValue+3);
    			    				//System.out.println("FB In Blockade ID = " + agent.getID() + " Road ID = " + keyID);
    			    			}
    			    			break;
    			    		}
    		            }
    		        }
    		    }
    		}
	    }
   }

	private boolean isInBlockade(Human human) {
		int agentX = human.getX();
		int agentY = human.getY();
		StandardEntity positionEntity = this.worldInfo.getPosition(human);
		if(positionEntity instanceof Road){
		    Road road = (Road)positionEntity;
		    if(road.isBlockadesDefined() && road.getBlockades().size() > 0){
		    	for(Blockade blockade : worldInfo.getBlockades(road)){
		    		if(blockade.getShape().contains(agentX, agentY)){
		    			return true;
		    		}
		    	}
		    }
		}
		return false;
	}
	
	private void prioritizeByIntersections(){
		Set<EntityID> intersections = new HashSet<>();
		this.worldInfo.getEntitiesOfType(ROAD,HYDRANT).stream()
			.filter(se -> ((Road)se).getNeighbours().size() >= 4)
			.forEach(se -> {
				((Area)se).getNeighbours().stream()
					.map(id -> this.worldInfo.getEntity(id))
					.filter(nse -> nse instanceof Road)
					.forEach(road -> {
						for(EntityID nextID : ((Road)road).getNeighbours()){
	    					StandardEntity nextNeighbour = this.worldInfo.getEntity(nextID);
	    					if(nextNeighbour instanceof Road && ((Road)nextNeighbour).getNeighbours().size() == 2){
    							intersections.add(road.getID());
	    					}
	    				}
					});
			});
		intersections.stream()
			.forEach(id -> {
				if(targetPriorities.containsKey(id)){
					double tempValue = targetPriorities.get(id);
    				targetPriorities.put(id, tempValue+1);
    				//System.out.println(" Road ID = " + id + " is Intersection");
				}
    		});
	}
	
    private void initPriorities(){
    	worldInfo.getEntitiesOfType(ROAD,HYDRANT).stream()
    		.forEach(se -> {
    			targetPriorities.put(se.getID(), 0.0);
    		});
    }
    
    private void deprioritizeByBlockade() {
    	for(EntityID id : this.worldInfo.getChanged().getChangedEntities()) {
            StandardEntity entity = this.worldInfo.getEntity(id);
            if(entity instanceof Road) {
                Road road = (Road)entity;
                if(!road.isBlockadesDefined() || road.getBlockades().isEmpty()) {
                	if(targetPriorities.containsKey(id)){
        				targetPriorities.put(id, -1.0);
        				//System.out.println(" Road ID = " + id + " has no blockade");
    				}
                }
            }
        }
	}
    
    private void prioritizeByCluster() {
    	int index = this.clustering.getClusterIndex(this.agentInfo.me().getID());
    	this.clustering.getClusterEntities(index).stream()
    		.filter(se -> se instanceof Road)
    		.map(se -> se.getID())
    		.forEach(id -> {
    			if(targetPriorities.containsKey(id)){
					double tempValue = targetPriorities.get(id);
    				targetPriorities.put(id, tempValue+2);
    				//System.out.println("Index " + index + " Road ID = " + id + " is in my cluster");
				}
    		});
	}
    
    private void prioritizeByRefuge() {
    	this.worldInfo.getEntitiesOfType(REFUGE).stream()
    		.map(se -> ((Area)se).getNeighbours())
    		.forEach(set -> {
	    		set.stream()
	    			.filter(ne -> worldInfo.getEntity(ne) instanceof Road)
	    			.forEach(id -> {
	    				if(targetPriorities.containsKey(id)){
	    					double tempValue = targetPriorities.get(id);
	        				targetPriorities.put(id, tempValue+2);
	        			}
	    			});
    		});
	}
    
    private void deprioritizeByBuilding() {
    	List<EntityID> burntList = this.worldInfo.getEntitiesOfType(BUILDING).stream()
    		.filter(se -> ((Building)se).getFierynessEnum() == Fieryness.BURNT_OUT)
    		.map(se -> se.getID())
    		.collect(Collectors.toList());
    	
    	for (EntityID idBurnt : burntList) {
    		worldInfo.getObjectIDsInRange(idBurnt, 10000).stream()
    			.filter(id -> this.worldInfo.getEntity(id) instanceof Road)
    			.forEach(id -> {
    				if(targetPriorities.containsKey(id)){
    					double tempValue = targetPriorities.get(id);
        				targetPriorities.put(id, tempValue-2);
        				//System.out.println(" Road ID = " + id + " is near BURNT_OUT");
        			}
    			});
    	}
    	//林追加部分
    	List<EntityID> heatingList = this.worldInfo.getEntitiesOfType(BUILDING).stream()
        		.filter(se -> ((Building)se).getFierynessEnum() == Fieryness.HEATING)
        		.map(se -> se.getID())
        		.collect(Collectors.toList());
    	heatingList.removeAll(burntList);
    	for (EntityID idHeating : heatingList) {
    		worldInfo.getObjectIDsInRange(idHeating, 10000).stream()
    			.filter(id -> this.worldInfo.getEntity(id) instanceof Road)
    			.forEach(id -> {
    				if(targetPriorities.containsKey(id)){
    					double tempValue = targetPriorities.get(id);
        				targetPriorities.put(id, tempValue+2);
        			}
    			});
    	}
	}
    
    //林追加部分
    private void deprioritizeByVisited() {
    	if(this.agentInfo.getTime() % 50 == 0){
    		this.visitedIds.clear();
    		return;
    	}
    	for(EntityID id : this.targetPriorities.keySet()){
    		if(this.visitedIds.contains(id)){
    			this.targetPriorities.put(id, 0.0);
    		}
    	}
	}
    
    private void prioritizeByMessage(MessageRoad messageRoad, Collection<EntityID> changedEntities) {
        if(messageRoad.isBlockadeDefined() && !changedEntities.contains(messageRoad.getBlockadeID())) {
            MessageUtil.reflectMessage(this.worldInfo, messageRoad);
        }
        if(messageRoad.isPassable()) {
			targetPriorities.put(messageRoad.getRoadID(), -1.0);
			//System.out.println(messageRoad.getRoadID() + " is Passable");
        }
    }
    
    private void prioritizeByMessage(MessageAmbulanceTeam messageAmbulanceTeam) {
        if(messageAmbulanceTeam.getPosition() == null) {
            return;
        }
    	// getAction()がACTION_RESCUEまたはACTION_LOADの場合，その建物のエントランスは優先度を下げる
        if(messageAmbulanceTeam.getAction() == MessageAmbulanceTeam.ACTION_RESCUE ||
       		messageAmbulanceTeam.getAction() == MessageAmbulanceTeam.ACTION_LOAD) {
            StandardEntity position = this.worldInfo.getEntity(messageAmbulanceTeam.getPosition());
            if(position != null && position instanceof Building) {
            	((Building)position).getNeighbours().stream()
            		.filter(id -> targetPriorities.containsKey(id))
            		.forEach(id -> {
            			double tempValue = targetPriorities.get(id);
                        targetPriorities.put(id, tempValue-2);
            		});
            }
        }
        // getAction()がACTION_MOVEの場合
    	// 1.targetが建物ならば，そのエントランスは優先度を上げる
        // 2.targetがhumanであり，そこが建物ならば，そのエントランスは優先度を上げる
        if(messageAmbulanceTeam.getAction() == MessageAmbulanceTeam.ACTION_MOVE) {
        	if(messageAmbulanceTeam.getTargetID() == null) {
                return;
            }
            StandardEntity target = this.worldInfo.getEntity(messageAmbulanceTeam.getTargetID());
            if(target instanceof Building) {
            	((Building)target).getNeighbours().stream()
	        		.filter(id -> targetPriorities.containsKey(id))
	        		.forEach(id -> {
	        			double tempValue = targetPriorities.get(id);
	                    targetPriorities.put(id, tempValue+1);
	        		});
            } else if(target instanceof Human) {
                Human human = (Human)target;
                if(human.isPositionDefined()) {
                    StandardEntity position = this.worldInfo.getPosition(human);
                    if(position instanceof Building) {
                    	((Building)position).getNeighbours().stream()
	                		.filter(id -> targetPriorities.containsKey(id))
	                		.forEach(id -> {
	                			double tempValue = targetPriorities.get(id);
	                            targetPriorities.put(id, tempValue+1);
	                		});
                    }
                }
            }
        }
    }
    
    private void prioritizeByMessage(MessageFireBrigade messageFireBrigade) {
        if(messageFireBrigade.getTargetID() == null) {
            return;
        }
        // getAction()がACTION_MOVEの場合
    	// 1.targetが建物ならば，そのエントランスは優先度を上げる
        // 2.targetがroadならば，そのroadの優先度を上げる
        if(messageFireBrigade.getAction() == MessageFireBrigade.ACTION_MOVE) {
        	if(messageFireBrigade.getTargetID() == null) {
                return;
            }
            StandardEntity target = this.worldInfo.getEntity(messageFireBrigade.getTargetID());
            if(target instanceof Road){
            	if(targetPriorities.containsKey(target.getID())){
	            	double tempValue = targetPriorities.get(target.getID());
	                targetPriorities.put(target.getID(), tempValue+3);
            	}
            }else if(target instanceof Building) {
            	((Building)target).getNeighbours().stream()
	        		.filter(id -> targetPriorities.containsKey(id))
	        		.forEach(id -> {
	        			double tempValue = targetPriorities.get(id);
	                    targetPriorities.put(id, tempValue+3);
	        		});
            }
        }
        // getAction()がACTION_REFILLの場合（補給地点に辿り着いているので）
    	// 1.targetが建物ならば，そのエントランスは優先度を下げる
        // 2.targetがroadならば，そのroadの優先度を下げる
        if(messageFireBrigade.getAction() == MessageFireBrigade.ACTION_REFILL) {
            StandardEntity target = this.worldInfo.getEntity(messageFireBrigade.getTargetID());
            if(target instanceof Building) {
            	((Building)target).getNeighbours().stream()
	        		.filter(id -> targetPriorities.containsKey(id))
	        		.forEach(id -> {
	        			double tempValue = targetPriorities.get(id);
	                    targetPriorities.put(id, tempValue-2);
	        		});
            } else if(target.getStandardURN() == HYDRANT) {
            	double tempValue = targetPriorities.get(target.getID());
                targetPriorities.put(target.getID(), tempValue-2);
            }
        }
    }
    
    private void prioritizeByMessage(MessagePoliceForce messagePoliceForce) {
        if(messagePoliceForce.getAction() == MessagePoliceForce.ACTION_CLEAR) {
            if(messagePoliceForce.getAgentID().getValue() != this.agentInfo.getID().getValue()) {
                if (messagePoliceForce.isTargetDefined()) {
                    EntityID targetID = messagePoliceForce.getTargetID();
                    if(targetID == null) {
                        return;
                    }
                    StandardEntity entity = this.worldInfo.getEntity(targetID);
                    if (entity == null) {
                        return;
                    }

                    if (entity instanceof Area) {
                    	if(targetPriorities.containsKey(targetID)){
                    		double tempValue = targetPriorities.get(targetID);
    	                    targetPriorities.put(targetID, tempValue-3);
                    	}
                        if(this.result != null && this.result.getValue() == targetID.getValue()) {
                        	// どちらか一方のresultをnullにする
                            if(this.agentInfo.getID().getValue() < messagePoliceForce.getAgentID().getValue()) {
                                this.result = null;
                            }
                        }
                    } else if (entity.getStandardURN() == BLOCKADE) {
                        EntityID position = ((Blockade) entity).getPosition();
                        double tempValue = targetPriorities.get(position);
	                    targetPriorities.put(position, tempValue-3);
                        if(this.result != null && this.result.getValue() == position.getValue()) {
                        	// どちらか一方のresultをnullにする
                            if(this.agentInfo.getID().getValue() < messagePoliceForce.getAgentID().getValue()) {
                                this.result = null;
                            }
                        }
                    }

                }
            }
        }
    }

    private void prioritizeByMessage(CommandPolice commandPolice) {
        boolean flag = false;
        if(commandPolice.isToIDDefined() && this.agentInfo.getID().getValue() == commandPolice.getToID().getValue()) {
            flag = true;
        } else if(commandPolice.isBroadcast()) {
            flag = true;
        }
        if(flag && commandPolice.getAction() == CommandPolice.ACTION_CLEAR) {
            if(commandPolice.getTargetID() == null) {
                return;
            }
            StandardEntity target = this.worldInfo.getEntity(commandPolice.getTargetID());
            if(target instanceof Area) {
            	if(targetPriorities.containsKey(target.getID())){
            		double tempValue = targetPriorities.get(target.getID());
                    targetPriorities.put(target.getID(), tempValue+3);
            	}
            } else if(target.getStandardURN() == BLOCKADE) {
                Blockade blockade = (Blockade)target;
                if(blockade.isPositionDefined()) {
                	if(targetPriorities.containsKey(blockade.getPosition())){
                		double tempValue = targetPriorities.get(blockade.getPosition());
	                    targetPriorities.put(blockade.getPosition(), tempValue+3);
                	}
                }
            }
        }
    }
    
    private Map<Building, Integer> crowdPosition(List<Human> list) {  
		Map<Building, Integer> pMap = new HashMap<Building, Integer>();  
		int k=0;  
		for(Human human : list){//サーバーから埋まっている人を持ってきて建物に何人いるか判定  
			StandardEntity pos = worldInfo.getEntity(human.getPosition()); 
			if(pos instanceof Building){
				Building posBuilding=(Building)pos;  
				if (pMap.containsKey(posBuilding)){//buildingの中にいるかを判定  
					k=pMap.get(posBuilding)+1;  
				}  
				pMap.put(posBuilding, k);
			}  
		}
		return pMap;
	}
    
    private void prioritizeByCrowdMember(){
    	Map<Building, Integer> crowdMember =  this.crowdPosition(this.buriedHumans);
 
    	for(Human human : this.buriedHumans){
			StandardEntity se = this.worldInfo.getEntity(human.getPosition());
			if (se instanceof Building) {
				Building building = (Building)se;
				if(!building.isOnFire()){
					int numHuman = crowdMember.containsKey(building) ? crowdMember.get(building) : 0;
					if(numHuman > 1){
						for(EntityID id : building.getNeighbours()){
							if(targetPriorities.containsKey(id)){
								double tempValue = targetPriorities.get(id);
			    			    targetPriorities.put(id, tempValue+1);
							}
				        }
					}
				}
			}
		}
    }
}
