package RIO.module.complex;

import adf.agent.communication.MessageManager;
import adf.agent.communication.standard.bundle.MessageUtil;
import adf.agent.communication.standard.bundle.information.*;
import adf.agent.communication.standard.bundle.centralized.CommandFire;
import adf.agent.communication.standard.bundle.centralized.CommandPolice;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.communication.CommunicationMessage;
import adf.component.module.algorithm.Clustering;
import adf.component.module.complex.HumanDetector;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.misc.geometry.Vector2D;
import rescuecore2.standard.entities.*;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.worldmodel.EntityID;

import java.util.*;

import static rescuecore2.standard.entities.StandardEntityURN.*;

public class RioneVictimDetector extends HumanDetector {
    private Clustering clustering;

    private EntityID result;
	private HashMap<EntityID, EntityID> results;
	
	private ArrayList<AmbulanceTeam> activeAT;
	private ArrayList<Human> targets;
	private ArrayList<AmbulanceTeam> workingAT = new ArrayList<>();
	private ArrayList<Human> workedtargets = new ArrayList<>();

    private int sendTime;
    private int sendingAvoidTimeClearRequest;

    private int moveDistance;
    private EntityID lastPosition;
    private int positionCount;

    public RioneVictimDetector(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData) {
        super(ai, wi, si, moduleManager, developData);

        this.result = null;
		this.results = new HashMap<>();
		
		this.activeAT = new ArrayList<>();
		this.targets = new ArrayList<>();
		this.workingAT = new ArrayList<>();
		this.workedtargets = new ArrayList<>();
		
        this.sendTime = 0;
        this.sendingAvoidTimeClearRequest = developData.getInteger("SampleHumanDetector.sendingAvoidTimeClearRequest", 5);
        this.moveDistance = developData.getInteger("SampleHumanDetector.moveDistance", 40000);

        switch  (scenarioInfo.getMode()) {
            case PRECOMPUTATION_PHASE:
                this.clustering = moduleManager.getModule("SampleHumanDetector.Clustering", "adf.sample.module.algorithm.SampleKMeans");
                break;
            case PRECOMPUTED:
                this.clustering = moduleManager.getModule("SampleHumanDetector.Clustering", "adf.sample.module.algorithm.SampleKMeans");
                break;
            case NON_PRECOMPUTE:
                this.clustering = moduleManager.getModule("SampleHumanDetector.Clustering", "adf.sample.module.algorithm.SampleKMeans");
                break;
        }
    }

    @Override
    public HumanDetector updateInfo(MessageManager messageManager) {
        super.updateInfo(messageManager);
        this.clustering.updateInfo(messageManager);
        sendChangedEntities(messageManager);
        receiveMessage(messageManager);
        selectTargets();
        askForClearAndExtinguish(messageManager);
        if(this.getCountUpdateInfo() >= 2) {
            return this;
        }
        return this;
    }
    
    private void sendChangedEntities(MessageManager messageManager) {
        //自分の視界情報を送信
        for(EntityID id : this.worldInfo.getChanged().getChangedEntities()) {
        	StandardEntity standardEntity = this.worldInfo.getEntity(id);
        	if(standardEntity instanceof Civilian) {
        		Civilian civilian = (Civilian)standardEntity;
        		StandardEntity position = this.worldInfo.getEntity(civilian.getPosition());
        		StandardEntityURN positionURN = position.getStandardURN();
        		if(positionURN != REFUGE){
        			if(civilian.isHPDefined() && civilian.getHP() > 0){
        				if((civilian.isBuriednessDefined() && civilian.getBuriedness() > 0) ||
        						(civilian.isDamageDefined() && civilian.getDamage() > 0)){
                			messageManager.addMessage(new MessageCivilian(true, civilian));
        				}
        			}
        		}
        	}else if(standardEntity instanceof Building){
        		Building building = (Building)standardEntity;
        		messageManager.addMessage(new MessageBuilding(true, building));
        	}else if(standardEntity instanceof Road){
        		Road road = (Road)standardEntity;
        		if(road.isBlockadesDefined() && road.getBlockades().size() > 0){
        			for(Blockade blockade : this.worldInfo.getBlockades(road)){
                		messageManager.addMessage(new MessageRoad(true, road, blockade, false, true));
                		break;
        			}
        		}else{
        			messageManager.addMessage(new MessageRoad(true, road, null, true, true));
        		}
        	}
        }
	}
    
    private void receiveMessage(MessageManager messageManager) {
    	this.workingAT.clear();
    	this.workedtargets.clear();
        //他Agentが送信した情報を受信して処理
        Set<EntityID> changedEntities = this.worldInfo.getChanged().getChangedEntities();
        for(CommunicationMessage message : messageManager.getReceivedMessageList()) {
            Class<? extends CommunicationMessage> messageClass = message.getClass();
            if(messageClass == MessageCivilian.class) {
                MessageCivilian mc = (MessageCivilian) message;
                if(!changedEntities.contains(mc.getAgentID())){
                    MessageUtil.reflectMessage(this.worldInfo, mc);
                }
            } else if(messageClass == MessageBuilding.class){
            	MessageBuilding mb = (MessageBuilding)message;
                if(!changedEntities.contains(mb.getBuildingID())) {
                    MessageUtil.reflectMessage(this.worldInfo, mb);
                }
            } else if(messageClass == MessageAmbulanceTeam.class) {
                MessageAmbulanceTeam mat = (MessageAmbulanceTeam)message;
                if(!changedEntities.contains(mat.getAgentID())) {
                    MessageUtil.reflectMessage(this.worldInfo, mat);
                }
                if(mat.getAction() == MessageAmbulanceTeam.ACTION_RESCUE || mat.getAction() == MessageAmbulanceTeam.ACTION_LOAD){
                	if(mat.getAgentID() != this.agentInfo.me().getID()){
                		this.workingAT.add((AmbulanceTeam)this.worldInfo.getEntity(mat.getAgentID()));
                		this.workedtargets.add((Human)this.worldInfo.getEntity(mat.getTargetID()));
                	}
                }else if(mat.getAction() == MessageAmbulanceTeam.ACTION_MOVE) {
                	if(mat.getAgentID() != this.agentInfo.me().getID()){
                		StandardEntity standardEntity = this.worldInfo.getEntity(mat.getTargetID());
                		if(standardEntity.getStandardURN() == REFUGE) {
                			this.workingAT.add((AmbulanceTeam)this.worldInfo.getEntity(mat.getAgentID()));
                		}
                	}
                }
            } else if(messageClass == MessageFireBrigade.class) {
                MessageFireBrigade mfb = (MessageFireBrigade) message;
                if(!changedEntities.contains(mfb.getAgentID())) {
                    MessageUtil.reflectMessage(this.worldInfo, mfb);
                }
            } else if(messageClass == MessagePoliceForce.class) {
                MessagePoliceForce mpf = (MessagePoliceForce) message;
                if(!changedEntities.contains(mpf.getAgentID())) {
                    MessageUtil.reflectMessage(this.worldInfo, mpf);
                }
            }
        }
	}
    
    private void selectTargets() {
        //targetの絞り込み
        this.activeAT.clear();
        this.targets.clear();
        this.results.clear();
        int clusterIndex = this.clustering.getClusterIndex(this.agentInfo.me().getID());
        Collection<StandardEntity> elements = this.clustering.getClusterEntities(clusterIndex);
        
        for (StandardEntity next : this.worldInfo.getEntitiesOfType(AMBULANCE_TEAM)){
        	Human h = (Human) next;
        	StandardEntity positionEntity = this.worldInfo.getPosition(h);
        	if(positionEntity != null && positionEntity instanceof Area){
        		if(elements.contains(positionEntity) || elements.contains(h)){
        			if (h.isHPDefined() && h.getHP() > 0){
        				if(h.isBuriednessDefined() && h.getBuriedness() > 0){
        					targets.add(h);
        				}else{
        					activeAT.add((AmbulanceTeam)next);
        				}
        			}
        		}
        	}
        }
        for (StandardEntity next : this.worldInfo.getEntitiesOfType(FIRE_BRIGADE, POLICE_FORCE)) {
        	Human h = (Human) next;
        	StandardEntity positionEntity = this.worldInfo.getPosition(h);
        	if(positionEntity != null) {
        		if(elements.contains(positionEntity) || elements.contains(h)) {
        			if(positionEntity instanceof Building) {
        				if (h.isHPDefined() && h.isBuriednessDefined() && h.getHP() > 0 && h.getBuriedness() > 0) {
        					targets.add(h);
        				}
        			}
        		}
        	}
        }
        for (StandardEntity next : this.worldInfo.getEntitiesOfType(CIVILIAN)) {
        	Human h = (Human) next;
        	StandardEntity positionEntity = this.worldInfo.getPosition(h);
        	StandardEntityURN positionURN = positionEntity.getStandardURN();
        	if(positionEntity != null && positionEntity instanceof Area) {
        		if (elements.contains(positionEntity) || elements.contains(h)) {
        			if (h.isHPDefined() && h.getHP() > 0) {
        				if ((h.isBuriednessDefined() && h.getBuriedness() > 0) ||
        						(h.isDamageDefined() && h.getDamage() > 0) &&
        						(positionURN != REFUGE) && positionURN != AMBULANCE_TEAM) {
        					targets.add(h);
        				}
        			}
        		}
        	}
        }
        this.activeAT.removeAll(this.workingAT);
        this.targets.removeAll(this.workedtargets);
	}
    
    private void askForClearAndExtinguish(MessageManager messageManager) {
        int currentTime = this.agentInfo.getTime();
        Human agent = (Human)this.agentInfo.me();
        int agentX = agent.getX();
        int agentY = agent.getY();
        StandardEntity positionEntity = this.worldInfo.getPosition(agent);
        if(positionEntity instanceof Road) {
            Road road = (Road)positionEntity;
            if(road.isBlockadesDefined() && road.getBlockades().size() > 0) {
                for(Blockade blockade : this.worldInfo.getBlockades(road)) {
                    if(blockade == null || !blockade.isApexesDefined()) {
                        continue;
                    }
                    if(this.isInside(agentX, agentY, blockade.getApexes())) {
                        if ((this.sendTime + this.sendingAvoidTimeClearRequest) <= currentTime) {
                            this.sendTime = currentTime;
                            messageManager.addMessage(
                                    new CommandPolice(
                                            true,
                                            null,
                                            agent.getPosition(),
                                            CommandPolice.ACTION_CLEAR
                                    )
                            );
                            break;
                        }
                    }
                }
            }
            if(this.lastPosition != null && this.lastPosition.getValue() == road.getID().getValue()) {
                this.positionCount++;
                if(this.positionCount > this.getMaxTravelTime(road)) {
                    if ((this.sendTime + this.sendingAvoidTimeClearRequest) <= currentTime) {
                        this.sendTime = currentTime;
                        messageManager.addMessage(
                                new CommandPolice(
                                        true,
                                        null,
                                        agent.getPosition(),
                                        CommandPolice.ACTION_CLEAR
                                )
                        );
                    }
                }
            } else {
                this.lastPosition = road.getID();
                this.positionCount = 0;
            }
        }

        if(this.result != null) {
            StandardEntity entity = this.worldInfo.getEntity(this.result);
            if(entity != null && entity instanceof Human) {
                Human h = (Human)entity;
                if(h.isPositionDefined()) {
                    StandardEntity humanPosition = this.worldInfo.getPosition(h);
                    if(humanPosition != null) {
                        if (humanPosition instanceof Building && ((Building) humanPosition).isOnFire()) {
                            messageManager.addMessage(
                                    new CommandFire(
                                            true,
                                            null,
                                            humanPosition.getID(),
                                            CommandFire.ACTION_EXTINGUISH
                                    )
                            );
                        } else if (humanPosition.getStandardURN() == AMBULANCE_TEAM && h.getStandardURN() == CIVILIAN) {
                            messageManager.addMessage(
                                    new MessageCivilian(
                                            true,
                                            (Civilian) h
                                    )
                            );
                        }
                    }
                }
            }
        }
	}

    @Override
    public HumanDetector calc() {
        Human transportHuman =  this.agentInfo.someoneOnBoard();
        if(transportHuman != null) {
            this.result = transportHuman.getID();
            return this;
        }
        //targetをリセットするか判断する//でたりはいったり問題
        if(this.result != null) {
            Human target = (Human) this.worldInfo.getEntity(this.result);
            if (!target.isHPDefined() || target.getHP() == 0) {
                this.result = null;
            } else if(!target.isPositionDefined()) {
                this.result = null;
            } else {
                StandardEntity position = this.worldInfo.getPosition(target);
                StandardEntityURN positionURN = position.getStandardURN();
                if (positionURN == REFUGE || positionURN == AMBULANCE_TEAM) {
                    this.result = null;
                }else if(positionURN == BUILDING && ((Building)position).isOnFire() &&
        		(((Building)position).getFierynessEnum() == Fieryness.INFERNO || ((Building)position).getFierynessEnum() == Fieryness.BURNING)){
                	this.result = null;
                }
            }
        }           
        
        if(this.result == null) {
        	this.result = this.calcGaleShapley();
        }
        if(this.result == null){
        	this.result = this.calcGreedy();
        }
        return this;
        
    }
    
	private EntityID calcGaleShapley() {
		MatchingData data[][] = new MatchingData[this.activeAT.size()][this.targets.size()];
		MatchingData dataT[][] = new MatchingData[this.targets.size()][this.activeAT.size()]; //dataの転置行列を格納
		HashMap<EntityID, List<EntityID>> ATtoTarget = new HashMap<>();
		HashMap<EntityID, List<EntityID>> TargettoAT = new HashMap<>();
		Map<Building, Integer> croudMember =  this.croudedPosition(this.targets);
        
        if(this.activeAT.isEmpty() || this.targets.isEmpty()){
        	return null;
        }
		
		//優先度行列の計算
		for(int i=0;i<this.activeAT.size();i++){
			EntityID AT = this.activeAT.get(i).getID();
			for(int j=0;j<this.targets.size();j++){
				Human target = this.targets.get(j);
				int triage = this.triage(target, AT);
				if(triage <= 0){
					data[i][j] = new MatchingData(AT, target.getID(),10.0);
				}else{
					data[i][j] = new MatchingData(AT, target.getID(), triage/300.0); //正規化
					StandardEntity se = this.worldInfo.getEntity(target.getPosition());
					if (se instanceof Building) {
						Building building = (Building)se;
						if(building.isOnFire()){
							data[i][j].setScore(10.0);
						} else{
							int num_of_hum = croudMember.containsKey(building) ? croudMember.get(building) : 1;
							data[i][j].diffScore((num_of_hum-1)/4.0); //正規化(MAX値を5,MIN値を1と仮定)		
						}		
					}
				}
			}
		}
		
		//転置行列の作成
		for(int i=0;i<this.activeAT.size();i++){
			for(int j=0;j<this.targets.size();j++){
				dataT[j][i] = data[i][j];
			}
		}
		
		//AT→Target方向の優先度リストを作成
		for(int i=0;i<this.activeAT.size();i++){
			if(data[i].length > 0){
				List<MatchingData> list = Arrays.asList(data[i]);
				Collections.sort(list);
				ArrayList<EntityID> targetlist = new ArrayList<>();
				for(MatchingData matchingData : list){
					targetlist.add(matchingData.getTargetID());
				}
				ATtoTarget.put(this.activeAT.get(i).getID(), targetlist);
				targetlist = null;
			}else{
				continue;
			}
		}
		
		//Target→AT方向の優先度リストを作成
		for(int i=0;i<this.targets.size();i++){
			if(dataT[i].length > 0){
				List<MatchingData> list = Arrays.asList(dataT[i]);
				Collections.sort(list);
				ArrayList<EntityID> ATlist = new ArrayList<>();
				for(MatchingData matchingData : list){
					ATlist.add(matchingData.getATID());
				}
				TargettoAT.put(this.targets.get(i).getID(), ATlist);
				ATlist = null;
			}else{
				continue;
			}
		}
		
		//安定結婚問題を用いたマッチング
		//ATのほうが少ないとき
		if(this.activeAT.size() <= this.targets.size()){
			HashMap<AmbulanceTeam, Integer> countAT= new HashMap<>();
			for(AmbulanceTeam ambulanceTeam : this.activeAT){
				countAT.put(ambulanceTeam, 0);
			}
			while(!(this.activeAT.isEmpty())){
				AmbulanceTeam ambulanceTeam = this.activeAT.get(0);
				EntityID atid= ambulanceTeam.getID();
				if(countAT.get(ambulanceTeam).intValue() >= ATtoTarget.get(atid).size()){
					continue;
				}
				EntityID targetid = ATtoTarget.get(atid).get(countAT.get(ambulanceTeam));
				if(this.results.containsKey(targetid)){ //マッチング相手が既にいる場合
					List<EntityID> list = TargettoAT.get(targetid);
					if(list.indexOf(atid) < list.indexOf(results.get(targetid))){
						this.activeAT.add((AmbulanceTeam)this.worldInfo.getEntity(results.get(targetid)));
						this.results.put(targetid, atid);
						this.activeAT.remove(ambulanceTeam);
					}else{
						this.activeAT.remove(ambulanceTeam);
						this.activeAT.add(ambulanceTeam);
					}
				}else{
					results.put(targetid, atid);
					this.activeAT.remove(ambulanceTeam);
				}
				countAT.put(ambulanceTeam, countAT.get(ambulanceTeam)+1);
			}
			for(EntityID targetid : results.keySet()){
				if(results.get(targetid) == this.agentInfo.me().getID()){
					return targetid;
				}
			}
			return null;
		//Targetのほうが少ないとき（主にシミュレーション開始時）
		}else{
			HashMap<Human, Integer> counttargets = new HashMap<>();
			for(Human human : this.targets){
				counttargets.put(human, 0);
			}
			while(!(this.targets.isEmpty())){
				Human human = this.targets.get(0);
				EntityID targetid = human.getID();
				if(counttargets.get(human).intValue() >= TargettoAT.get(targetid).size()){
					continue;
				}
				EntityID atid = TargettoAT.get(targetid).get(counttargets.get(human));
				if(this.results.containsKey(atid)){ //マッチング相手が既にいる場合
					List<EntityID> list = ATtoTarget.get(atid);
					if(list.indexOf(targetid) < list.indexOf(this.results.get(atid))){
						this.targets.add((Human)this.worldInfo.getEntity(this.results.get(atid)));
						this.results.put(atid, targetid);
						this.targets.remove(human);
					}else{
						this.targets.remove(human);
						this.targets.add(human);
					}
				}else{
					this.results.put(atid, targetid);
					this.targets.remove(human);
				}
				counttargets.put(human, counttargets.get(human)+1);
			}
			if(this.results.containsKey(this.agentInfo.me().getID())){
				return this.results.get(this.agentInfo.me().getID());
			}
			return null;
		}
	}
	
	private EntityID calcGreedy() {
		ArrayList<MatchingData> results = new ArrayList<MatchingData>();
		Map<Building, Integer> croudMember = this.croudedPosition(this.targets);
		int count = 0;
       
		//評価値付	
		for(int i=0;i<this.targets.size();i++){
			Human human = this.targets.get(i);
			int triage = this.triage(human, this.agentInfo.getID());
			if(triage <= 0){
				continue;
			}else{
				results.add(new MatchingData(null, human.getID(), triage/300.0));
				StandardEntity se = this.worldInfo.getEntity(human.getPosition());
				if (se instanceof Building) {
					Building building = (Building)se;
					if(building.isOnFire()){
						results.get(count).setScore(10.0);
					} else{
						int num_of_hum = croudMember.get(building);
						results.get(count).diffScore((num_of_hum-1)/4.0); //正規化(MAX値を5,MIN値を1と仮定)		
					}
				}
				count++;
			}
		}
		
		if(results.size() > 0){
			return Collections.min(results).getTargetID();
		}
		return null;
	}
	
	private int triage(Human human, EntityID id) {
		Collection<StandardEntity> refuges = this.worldInfo.getEntitiesOfType(StandardEntityURN.REFUGE);
		if (refuges.isEmpty()) return 0;
		Refuge refuge = (Refuge) Collections.min(refuges, new DistanceSorter(this.worldInfo, this.worldInfo.getEntity(id)));
	
		if (human.isDamageDefined() && human.isHPDefined()) {
			if (human.getDamage() == 0) return 300;//ダメージがない場合、十分な(放置可能)サイクルを返す
			int SurvivalTime = (human.getHP()-100) / human.getDamage() ;
			int RescueTime = (this.worldInfo.getDistance(id, human.getID()) + human.getBuriedness() + this.worldInfo.getDistance(human.getID(), refuge.getID())) / 30000;
			return SurvivalTime - RescueTime;
		}
		return 0;//判定できない場合、放置可能サイクルを0で返す
	}
	
	private Map<Building, Integer> croudedPosition(List<Human> list) 
	{  
		Map<Building, Integer> pMap = new HashMap<Building, Integer>();  
		int k=0;  
		for(Human human : list)  
		{//サーバーから埋まっている人を持ってきて建物いるかを判定何人くらいいるのか  
			StandardEntity pos = worldInfo.getEntity(human.getPosition()); 
			if(pos instanceof Building){
				Building posBuilding=(Building)pos;  
				if (pMap.containsKey(posBuilding))  
				{//buildingの中にいるかを判定  
					k=pMap.get(posBuilding)+1;  
				}  
				pMap.put(posBuilding, k);  
         
			}  
		}
		return pMap;
	}
	
	private class MatchingData implements Comparable<MatchingData> {

		private EntityID AT;
		private EntityID target;
		private double score;
		
		public MatchingData(EntityID AT, EntityID target, double score) {
			this.AT = AT;
			this.target = target;
			this.score = score;
		}
		
		protected EntityID getATID() {
			return this.AT;
		}
		
		protected EntityID getTargetID() {
			return this.target;
		}
		
		protected void setScore(double score) {
			this.score = score;
		}
		
		protected void diffScore(double score) {
			this.score -= score;
		}
		
		public int compareTo(MatchingData machingData) {
			if(this.score < machingData.score){
				return -1;
			} else if(this.score > machingData.score){
				return 1;
			} else{
				return 0;
			}
		}
	}

    @Override
    public EntityID getTarget() {
        return this.result;
    }

    @Override
    public HumanDetector precompute(PrecomputeData precomputeData) {
        super.precompute(precomputeData);
        if(this.getCountPrecompute() >= 2) {
            return this;
        }
        this.clustering.precompute(precomputeData);
        return this;
    }

    @Override
    public HumanDetector resume(PrecomputeData precomputeData) {
        super.resume(precomputeData);
        if(this.getCountResume() >= 2) {
            return this;
        }
        this.clustering.resume(precomputeData);
        return this;
    }

    @Override
    public HumanDetector preparate() {
        super.preparate();
        if(this.getCountPreparate() >= 2) {
            return this;
        }
        this.clustering.preparate();
        return this;
    }

    private boolean isInside(double pX, double pY, int[] apex) {
        Point2D p = new Point2D(pX, pY);
        Vector2D v1 = (new Point2D(apex[apex.length - 2], apex[apex.length - 1])).minus(p);
        Vector2D v2 = (new Point2D(apex[0], apex[1])).minus(p);
        double theta = this.getAngle(v1, v2);

        for(int i = 0; i < apex.length - 2; i += 2) {
            v1 = (new Point2D(apex[i], apex[i + 1])).minus(p);
            v2 = (new Point2D(apex[i + 2], apex[i + 3])).minus(p);
            theta += this.getAngle(v1, v2);
        }
        return Math.round(Math.abs((theta / 2) / Math.PI)) >= 1;
    }

    private double getAngle(Vector2D v1, Vector2D v2) {
        double flag = (v1.getX() * v2.getY()) - (v1.getY() * v2.getX());
        double angle = Math.acos(((v1.getX() * v2.getX()) + (v1.getY() * v2.getY())) / (v1.getLength() * v2.getLength()));
        if(flag > 0) {
            return angle;
        }
        if(flag < 0) {
            return -1 * angle;
        }
        return 0.0D;
    }

    private class DistanceSorter implements Comparator<StandardEntity> {
        private StandardEntity reference;
        private WorldInfo worldInfo;

        DistanceSorter(WorldInfo wi, StandardEntity reference) {
            this.reference = reference;
            this.worldInfo = wi;
        }

        public int compare(StandardEntity a, StandardEntity b) {
            int d1 = this.worldInfo.getDistance(this.reference, a);
            int d2 = this.worldInfo.getDistance(this.reference, b);
            return d1 - d2;
        }
    }

    private int getMaxTravelTime(Area area) {
        int distance = 0;
        List<Edge> edges = new ArrayList<>();
        for(Edge edge : area.getEdges()) {
            if(edge.isPassable()) {
                edges.add(edge);
            }
        }
        if(edges.size() <= 1) {
            return Integer.MAX_VALUE;
        }
        for(int i = 0; i < edges.size(); i++) {
            for(int j = 0; j < edges.size(); j++) {
                if(i != j) {
                    Edge edge1 = edges.get(i);
                    double midX1 = (edge1.getStartX() + edge1.getEndX()) / 2;
                    double midY1 = (edge1.getStartY() + edge1.getEndY()) / 2;
                    Edge edge2 = edges.get(j);
                    double midX2 = (edge2.getStartX() + edge2.getEndX()) / 2;
                    double midY2 = (edge2.getStartY() + edge2.getEndY()) / 2;
                    int d = this.getDistance(midX1, midY1, midX2, midY2);
                    if(distance < d) {
                        distance = d;
                    }
                }
            }
        }
        if(distance > 0) {
            return (distance / this.moveDistance) + ((distance % this.moveDistance) > 0 ? 1 : 0) + 1;
        }
        return Integer.MAX_VALUE;
    }

    private int getDistance(double fromX, double fromY, double toX, double toY) {
        double dx = toX - fromX;
        double dy = toY - fromY;
        return (int)Math.hypot(dx, dy);
    }
}
